############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Jannis Gebauer <ja.geb@me.com>                                #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2025 Cristiano Salerno <119511125+csalerno-asml@users.noreply.github.com>#
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from . import Framework


class GitRef(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.ref = self.g.get_user().get_repo("PyGithub").get_git_ref("heads/BranchCreatedByPyGithub")

    def testAttributes(self):
        self.assertIsNone(self.ref.node_id)
        self.assertEqual(self.ref.object.sha, "1292bf0e22c796e91cc3d6e24b544aece8c21f2a")
        self.assertEqual(self.ref.object.type, "commit")
        self.assertEqual(
            self.ref.object.url,
            "https://api.github.com/repos/jacquev6/PyGithub/git/commits/1292bf0e22c796e91cc3d6e24b544aece8c21f2a",
        )
        self.assertEqual(self.ref.ref, "ref/heads/BranchCreatedByPyGithub")
        self.assertEqual(
            self.ref.url,
            "https://api.github.com/repos/jacquev6/PyGithub/git/ref/heads/BranchCreatedByPyGithub",
        )

        self.assertEqual(repr(self.ref), 'GitRef(ref="ref/heads/BranchCreatedByPyGithub")')
        self.assertEqual(
            repr(self.ref.object),
            'GitObject(sha="1292bf0e22c796e91cc3d6e24b544aece8c21f2a")',
        )

    def testLazyAttributes(self):
        ref = self.g.withLazy(True).get_repo("lazy/repo").get_git_ref("refs/heads/main")
        self.assertEqual(str(ref), 'GitRef(ref="refs/heads/main")')
        self.assertEqual(ref.ref, "refs/heads/main")
        self.assertEqual(ref.url, "/repos/lazy/repo/git/ref/refs/heads/main")

    def testEdit(self):
        self.ref.edit("04cde900a0775b51f762735637bd30de392a2793")

    def testEditWithForce(self):
        self.ref.edit("4303c5b90e2216d927155e9609436ccb8984c495", force=True)

    def testDelete(self):
        self.ref.delete()



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 Agor Maxime <maxime.agor23@gmail.com>                         #
# Copyright 2018 Joel Koglin <JoelKoglin@gmail.com>                            #
# Copyright 2018 Shubham Singh <41840111+singh811@users.noreply.github.com>    #
# Copyright 2018 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2018 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2018 h.shi <10385628+AnYeMoWang@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2021 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from . import Framework


class Search(Framework.TestCase):
    def setUp(self):
        super().setUp()

    def testSearchUsers(self):
        users = self.g.search_users("vincent", sort="followers", order="desc")
        self.assertEqual(users.totalCount, 2781)

    def testPaginateSearchUsers(self):
        users = self.g.search_users("", location="Berlin")
        self.assertListKeyBegin(
            users,
            lambda u: u.login,
            [
                "cloudhead",
                "felixge",
                "sferik",
                "rkh",
                "jezdez",
                "janl",
                "marijnh",
                "nikic",
                "igorw",
                "froschi",
                "svenfuchs",
                "omz",
                "chad",
                "bergie",
                "roidrage",
                "pcalcado",
                "durran",
                "hukl",
                "mttkay",
                "aFarkas",
                "ole",
                "hagenburger",
                "jberkel",
                "naderman",
                "joshk",
                "pudo",
                "robb",
                "josephwilk",
                "hanshuebner",
                "txus",
                "paulasmuth",
                "splitbrain",
                "langalex",
                "bendiken",
                "stefanw",
            ],
        )
        self.assertEqual(users.totalCount, 6038)
        self.assertEqual(users[0].score, 1.0)

    def testGetPageOnSearchUsers(self):
        users = self.g.search_users("", location="Berlin")
        self.assertEqual(
            [u.login for u in users.get_page(7)],
            [
                "ursachec",
                "bitboxer",
                "fs111",
                "michenriksen",
                "witsch",
                "booo",
                "mortice",
                "r0man",
                "MikeBild",
                "mhagger",
                "bkw",
                "fwbrasil",
                "mschneider",
                "lydiapintscher",
                "asksven",
                "iamtimm",
                "sneak",
                "kr1sp1n",
                "Feh",
                "GordonLesti",
                "annismckenzie",
                "eskimoblood",
                "tsujigiri",
                "riethmayer",
                "lauritzthamsen",
                "scotchi",
                "peritor",
                "toto",
                "hwaxxer",
                "lukaszklis",
            ],
        )

    def testSearchRepos(self):
        repos = self.g.search_repositories("github", sort="stars", order="desc", language="Python")
        self.assertListKeyBegin(
            repos,
            lambda r: r.full_name,
            [
                "kennethreitz/legit",
                "RuudBurger/CouchPotatoV1",
                "gelstudios/gitfiti",
                "gpjt/webgl-lessons",
                "jacquev6/PyGithub",
                "aaasen/github_globe",
                "hmason/gitmarks",
                "dnerdy/factory_boy",
                "binaryage/drydrop",
                "bgreenlee/sublime-github",
                "karan/HackerNewsAPI",
                "mfenniak/pyPdf",
                "skazhy/github-decorator",
                "llvmpy/llvmpy",
                "lexrupy/gmate",
                "ask/python-github2",
                "audreyr/cookiecutter-pypackage",
                "tabo/django-treebeard",
                "dbr/tvdb_api",
                "jchris/couchapp",
                "joeyespo/grip",
                "nigelsmall/py2neo",
                "ask/chishop",
                "sigmavirus24/github3.py",
                "jsmits/github-cli",
                "lincolnloop/django-layout",
                "amccloud/django-project-skel",
                "Stiivi/brewery",
                "webpy/webpy.github.com",
                "dustin/py-github",
                "logsol/Github-Auto-Deploy",
                "cloudkick/libcloud",
                "berkerpeksag/github-badge",
                "bitprophet/ssh",
                "azavea/OpenTreeMap",
            ],
        )

    def testSearchReposWithNoResults(self):
        repos = self.g.search_repositories("doesnotexist")
        self.assertEqual(repos.totalCount, 0)

    def testSearchIssues(self):
        issues = self.g.search_issues("compile", sort="comments", order="desc", language="C++")
        self.assertListKeyBegin(
            issues,
            lambda i: i.id,
            [
                12068673,
                23250111,
                14371957,
                9423897,
                24277400,
                2408877,
                11338741,
                13980502,
                27697165,
                23102422,
            ],
        )
        self.assertEqual(issues[0].score, 0.08252439)

    def testSearchCommits(self):
        pages = self.g.search_commits(query="hash:5b0224e868cc9242c9450ef02efbe3097abd7ba2")
        commits = list(pages)
        self.assertEqual(pages.totalCount, 12)
        self.assertEqual(commits[0].commit.message, "Fix README instructions")
        self.assertEqual(commits[0].score, 1.0)
        self.assertEqual(commits[0].sha, "5b0224e868cc9242c9450ef02efbe3097abd7ba2")

    def testSearchCommitsOrder(self):
        pages = self.g.search_commits(
            query="hash:1265747e992ba7d34a469b6b2f527809f8bf7067",
            sort="author-date",
            order="asc",
            merge="false",
        )
        commits = list(pages)
        self.assertEqual(pages.totalCount, 4)
        self.assertEqual(len(commits[0].commit.message), 490)
        self.assertEqual(commits[0].score, 1.0)
        self.assertEqual(commits[0].sha, "1265747e992ba7d34a469b6b2f527809f8bf7067")

    def testSearchTopics(self):
        topics = self.g.search_topics("python", repositories=">950")
        self.assertListKeyBegin(
            topics,
            lambda r: r.name,
            ["python", "django", "flask", "ruby", "scikit-learn", "wagtail"],
        )

    def testPaginateSearchTopics(self):
        repos = self.g.search_topics("python", repositories=">950")
        self.assertEqual(repos.totalCount, 6)

    def testSearchCode(self):
        files = self.g.search_code("toto", sort="indexed", order="asc", user="jacquev6")
        self.assertListKeyEqual(
            files,
            lambda f: f.name,
            [
                "Commit.setUp.txt",
                "PullRequest.testGetFiles.txt",
                "NamedUser.testGetEvents.txt",
                "PullRequest.testCreateComment.txt",
                "PullRequestFile.setUp.txt",
                "Repository.testGetIssuesWithWildcards.txt",
                "Repository.testGetIssuesWithArguments.txt",
                "test_ebnf.cpp",
                "test_abnf.cpp",
                "PullRequestFile.py",
                "SystemCalls.py",
                "tests.py",
                "LexerTestCase.py",
                "ParserTestCase.py",
            ],
        )
        self.assertEqual(files[0].score, 0.31651077)
        self.assertEqual(files[0].repository.full_name, "jacquev6/PyGithub")
        content = files[0].decoded_content
        if isinstance(content, bytes):
            content = content.decode("utf-8")
        self.assertEqual(content[:30], "https\nGET\napi.github.com\nNone\n")

    def testSearchHighlightingCode(self):
        files = self.g.search_code("toto", sort="indexed", order="asc", user="jacquev6", highlight=True)
        self.assertEqual(files[0].score, 14.030813)
        self.assertEqual(
            files[0].text_matches,
            [
                {
                    "fragment": ".assertEqual(\n"
                    "            self.recorded.instance_method(42, 43, 44, 45, "
                    "toto=46, tutu=47",
                    "matches": [{"indices": [72, 76], "text": "toto"}],
                    "object_type": "FileContent",
                    "object_url": "https://api.github.com/repositories/6430524/contents/MockMockMock/tests/record_replay.py?ref=562a55542f55426f6853f3013309c85f402c359e",
                    "property": "content",
                },
                {
                    "fragment": "),\n"
                    "            \"(42, 43, (44, 45), [('toto', 46), ('tutu', "
                    '47)])"\n'
                    "        )\n"
                    "        self.assertEqual",
                    "matches": [{"indices": [38, 42], "text": "toto"}],
                    "object_type": "FileContent",
                    "object_url": "https://api.github.com/repositories/6430524/contents/MockMockMock/tests/record_replay.py?ref=562a55542f55426f6853f3013309c85f402c359e",
                    "property": "content",
                },
            ],
        )

    def testUrlquotingOfQualifiers(self):
        # Example taken from #236
        issues = self.g.search_issues("repo:saltstack/salt-api type:Issues", updated=">2014-03-04T18:28:11Z")
        self.assertEqual(issues[0].id, 29138794)

    def testUrlquotingOfQuery(self):
        # Example taken from #236
        issues = self.g.search_issues("repo:saltstack/salt-api type:Issues updated:>2014-03-04T18:28:11Z")
        self.assertEqual(issues[0].id, 29138794)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 YugoHino <henom06@gmail.com>                                  #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2026 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################


from . import Framework


class EnterpriseAdmin(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.enterprise = self.g.get_enterprise("beaver-group")

    def testAttributes(self):
        self.assertEqual(self.enterprise.slug, "beaver-group")
        self.assertEqual(self.enterprise.url, "/enterprises/beaver-group")
        self.assertEqual(repr(self.enterprise), 'Enterprise(slug="beaver-group")')

    def testGetConsumedLicenses(self):
        consumed_licenses = self.enterprise.get_consumed_licenses()
        self.assertEqual(consumed_licenses.total_seats_consumed, 102)
        self.assertEqual(consumed_licenses.total_seats_purchased, 103)

    def testGetEnterpriseUsers(self):
        enterprise_users = self.enterprise.get_consumed_licenses().users
        enterprise_users_list = [
            [
                users.github_com_login,
                users.github_com_name,
                users.enterprise_server_user_ids,
                users.github_com_user,
                users.enterprise_server_user,
                users.visual_studio_subscription_user,
                users.license_type,
                users.github_com_profile,
                users.github_com_member_roles,
                users.github_com_enterprise_roles,
                users.github_com_verified_domain_emails,
                users.github_com_saml_name_id,
                users.github_com_orgs_with_pending_invites,
                users.github_com_two_factor_auth,
                users.enterprise_server_primary_emails,
                users.visual_studio_license_status,
                users.visual_studio_subscription_email,
                users.total_user_accounts,
            ]
            for users in enterprise_users
        ]
        self.assertEqual(len(enterprise_users_list), 102)
        self.assertEqual(enterprise_users_list[42][0], "beaver-user043")

    def testGetEnterpriseGetUsers(self):
        enterprise_users = self.enterprise.get_consumed_licenses(licence_users_per_page=1).get_users(
            licence_users_per_page=100
        )
        enterprise_users_list = list(enterprise_users)
        self.assertEqual(len(enterprise_users_list), 102)
        self.assertEqual(enterprise_users_list[42].github_com_login, "beaver-user043")



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Joseph Henrich <crimsonknave@gmail.com>                       #
# Copyright 2024 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2024 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from datetime import datetime, timezone
from decimal import Decimal

import github.GlobalAdvisory

from . import Framework


class GlobalAdvisory(Framework.TestCase):
    advisory: github.GlobalAdvisory.GlobalAdvisory

    def setUp(self):
        super().setUp()

    def testAttributes(self):
        self.advisory = self.g.get_global_advisory("GHSA-wqc8-x2pr-7jqh")
        self.assertListKeyEqual(
            self.advisory.credits,
            lambda e: (e.user.login, e.type),
            [
                ("loechel", "remediation_developer"),
                ("Quasar0147", "reporter"),
                ("despawningbone", "reporter"),
                ("dataflake", "coordinator"),
                ("nneonneo", "other"),
            ],
        )
        self.assertEqual(self.advisory.cve_id, "CVE-2023-37271")
        self.assertEqual(self.advisory.cvss.version, Decimal("3.1"))
        self.assertEqual(self.advisory.cvss.score, Decimal("8.4"))
        self.assertEqual(self.advisory.cvss.vector_string, "CVSS:3.1/AV:N/AC:H/PR:L/UI:N/S:C/C:H/I:H/A:L")
        self.assertEqual(
            self.advisory.cvss_severities,
            {
                "cvss_v3": {"vector_string": "CVSS:3.1/AV:N/AC:H/PR:L/UI:N/S:C/C:H/I:H/A:L", "score": 8.4},
                "cvss_v4": {"vector_string": None, "score": 0.0},
            },
        )
        self.assertListKeyEqual(
            self.advisory.cwes,
            lambda e: (e.cwe_id, e.name),
            [
                ("CWE-913", "Improper Control of Dynamically-Managed Code Resources"),
            ],
        )
        self.assertEqual(
            self.advisory.description[:100],
            "### Impact\n\nRestrictedPython does not check access to stack frames and their attributes. Stack frame",
        )
        self.assertEqual(self.advisory.epss, {"percentage": 0.00156, "percentile": 0.52841})
        self.assertEqual(self.advisory.ghsa_id, "GHSA-wqc8-x2pr-7jqh")
        self.assertEqual(self.advisory.github_reviewed_at, datetime(2023, 7, 10, 21, 53, 22, tzinfo=timezone.utc))
        self.assertEqual(
            self.advisory.html_url,
            "https://github.com/advisories/GHSA-wqc8-x2pr-7jqh",
        )
        self.assertListEqual(
            self.advisory.identifiers,
            [{"type": "GHSA", "value": "GHSA-wqc8-x2pr-7jqh"}, {"type": "CVE", "value": "CVE-2023-37271"}],
        )
        self.assertEqual(self.advisory.nvd_published_at, datetime(2023, 7, 11, 18, 15, 20, tzinfo=timezone.utc))
        self.assertEqual(
            self.advisory.published_at,
            datetime(2023, 7, 10, 21, 53, 22, tzinfo=timezone.utc),
        )
        self.assertListEqual(
            self.advisory.references,
            [
                "https://github.com/zopefoundation/RestrictedPython/security/advisories/GHSA-wqc8-x2pr-7jqh",
                "https://github.com/zopefoundation/RestrictedPython/commit/c8eca66ae49081f0016d2e1f094c3d72095ef531",
                "https://nvd.nist.gov/vuln/detail/CVE-2023-37271",
                "https://github.com/pypa/advisory-database/tree/main/vulns/restrictedpython/PYSEC-2023-118.yaml",
                "https://github.com/advisories/GHSA-wqc8-x2pr-7jqh",
            ],
        )
        self.assertEqual(
            self.advisory.repository_advisory_url,
            "https://api.github.com/repos/zopefoundation/RestrictedPython/security-advisories/GHSA-wqc8-x2pr-7jqh",
        )
        self.assertEqual(self.advisory.severity, "high")
        self.assertEqual(
            self.advisory.source_code_location,
            "https://github.com/zopefoundation/RestrictedPython",
        )
        self.assertEqual(
            self.advisory.summary,
            "RestrictedPython vulnerable to arbitrary code execution via stack frame sandbox escape",
        )
        self.assertEqual(
            self.advisory.type,
            "reviewed",
        )
        self.assertEqual(self.advisory.updated_at, datetime(2023, 11, 7, 5, 5, 13, tzinfo=timezone.utc))
        self.assertEqual(self.advisory.url, "https://api.github.com/advisories/GHSA-wqc8-x2pr-7jqh")
        self.assertListKeyEqual(
            self.advisory.vulnerabilities,
            lambda e: (
                (e.package.ecosystem, e.package.name),
                e.patched_versions,
                e.vulnerable_functions,
                e.vulnerable_version_range,
            ),
            [
                (("pip", "RestrictedPython"), None, [], "< 5.3"),
                (("pip", "RestrictedPython"), None, [], ">= 6.0a1.dev0, < 6.1"),
                (("pip", "restrictedpython"), None, [], ">= 0, < 5.3"),
            ],
        )
        self.assertEqual(self.advisory.withdrawn_at, None)

    def testNewlyReleased(self):
        """
        Test an advisory that was freshly released and does not have values for all fields.
        """
        self.advisory = self.g.get_global_advisory("GHSA-cx3j-qqxj-9597")
        self.assertListKeyEqual(
            self.advisory.credits,
            lambda e: (e.user.login, e.type),
            [],
        )
        self.assertEqual(self.advisory.cve_id, "CVE-2023-3481")
        self.assertEqual(self.advisory.cvss.version, None)
        self.assertEqual(self.advisory.cvss.score, None)
        self.assertEqual(self.advisory.cvss.vector_string, None)
        self.assertListKeyEqual(
            self.advisory.cwes,
            lambda e: (e.cwe_id, e.name),
            [
                ("CWE-80", "Improper Neutralization of Script-Related HTML Tags in a Web Page (Basic XSS)"),
                ("CWE-116", "Improper Encoding or Escaping of Output"),
            ],
        )
        self.assertEqual(
            self.advisory.description, "### Impact\nCritters version 0.0.17-0.0.19 have an issue when parsing..."
        )
        self.assertEqual(self.advisory.ghsa_id, "GHSA-cx3j-qqxj-9597")
        self.assertEqual(
            self.advisory.github_reviewed_at,
            datetime(2023, 8, 11, 18, 57, 53, tzinfo=timezone.utc),
        )
        self.assertEqual(
            self.advisory.html_url,
            "https://github.com/advisories/GHSA-cx3j-qqxj-9597",
        )
        self.assertListEqual(
            self.advisory.identifiers,
            [{"type": "GHSA", "value": "GHSA-cx3j-qqxj-9597"}, {"type": "CVE", "value": "CVE-2023-3481"}],
        )
        self.assertEqual(self.advisory.nvd_published_at, None)
        self.assertEqual(
            self.advisory.published_at,
            datetime(2023, 8, 11, 18, 57, 53, tzinfo=timezone.utc),
        )
        self.assertListEqual(
            self.advisory.references,
            [
                "https://github.com/GoogleChromeLabs/critters/security/advisories/GHSA-cx3j-qqxj-9597",
                "https://github.com/GoogleChromeLabs/critters/pull/133",
                "https://github.com/GoogleChromeLabs/critters/commit/7757902c9e0b3285d516359b3cb602cd9d50d80e",
                "https://github.com/advisories/GHSA-cx3j-qqxj-9597",
            ],
        )
        self.assertEqual(
            self.advisory.repository_advisory_url,
            "https://api.github.com/repos/GoogleChromeLabs/critters/security-advisories/GHSA-cx3j-qqxj-9597",
        )
        self.assertEqual(self.advisory.severity, "high")
        self.assertEqual(
            self.advisory.source_code_location,
            "https://github.com/GoogleChromeLabs/critters",
        )
        self.assertEqual(
            self.advisory.summary,
            "Critters Cross-site Scripting Vulnerability",
        )
        self.assertEqual(
            self.advisory.type,
            "reviewed",
        )
        self.assertEqual(
            self.advisory.updated_at,
            datetime(2023, 8, 11, 18, 57, 54, tzinfo=timezone.utc),
        )
        self.assertEqual(
            self.advisory.url,
            "https://api.github.com/advisories/GHSA-cx3j-qqxj-9597",
        )
        self.assertListKeyEqual(
            self.advisory.vulnerabilities,
            lambda e: (
                (e.package.ecosystem, e.package.name),
                e.patched_versions,
                e.vulnerable_functions,
                e.vulnerable_version_range,
            ),
            [(("npm", "critters"), None, [], ">= 0.0.17, <= 0.0.19")],
        )
        self.assertEqual(self.advisory.withdrawn_at, None)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 AKFish <akfish@gmail.com>                                     #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2015 Uriel Corfa <uriel@corfa.fr>                                  #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2017 Chris McBride <thehighlander@users.noreply.github.com>        #
# Copyright 2017 Hugo <hugovk@users.noreply.github.com>                        #
# Copyright 2017 Simon <spam@esemi.ru>                                         #
# Copyright 2018 Arda Kuyumcu <kuyumcuarda@gmail.com>                          #
# Copyright 2018 Jacopo Notarstefano <jacopo.notarstefano@gmail.com>           #
# Copyright 2018 Laurent Mazuel <lmazuel@microsoft.com>                        #
# Copyright 2018 Mike Miller <github@mikeage.net>                              #
# Copyright 2018 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Adam Baratz <adam.baratz@gmail.com>                           #
# Copyright 2019 Filipe Laíns <filipe.lains@gmail.com>                         #
# Copyright 2019 Isac Souza <isouza@daitan.com>                                #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Alice GIRARD <bouhahah@gmail.com>                             #
# Copyright 2020 Michał Górny <mgorny@gentoo.org>                              #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2021 Amador Pahim <apahim@redhat.com>                              #
# Copyright 2021 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Denis Blanchette <dblanchette@coveo.com>                      #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2023 Jonathan Leitschuh <jonathan.leitschuh@gmail.com>             #
# Copyright 2023 Trim21 <trim21.me@gmail.com>                                  #
# Copyright 2023 chantra <chantra@users.noreply.github.com>                    #
# Copyright 2025 Alex Olieman <alex@olieman.net>                               #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2025 Maja Massarini <2678400+majamassarini@users.noreply.github.com>#
# Copyright 2025 Matej Focko <mfocko@users.noreply.github.com>                 #
# Copyright 2025 Neel Malik <41765022+neel-m@users.noreply.github.com>         #
# Copyright 2026 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

import base64
import contextlib
import json
import os
import traceback
import unittest
import warnings
from collections.abc import Generator
from dataclasses import dataclass
from io import BytesIO
from typing import Any, Callable

import responses
from requests.structures import CaseInsensitiveDict
from urllib3.util import Url

import github
from github import Consts

APP_PRIVATE_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIICXAIBAAKBgQC+5ePolLv6VcWLp2f17g6r6vHl+eoLuodOOfUl8JK+MVmvXbPa
xDy0SS0pQhwTOMtB0VdSt++elklDCadeokhEoGDQp411o+kiOhzLxfakp/kewf4U
HJnu4M/A2nHmxXVe2lzYnZvZHX5BM4SJo5PGdr0Ue2JtSXoAtYr6qE9maQIDAQAB
AoGAFhOJ7sy8jG+837Clcihso+8QuHLVYTPaD+7d7dxLbBlS8NfaQ9Nr3cGUqm/N
xV9NCjiGa7d/y4w/vrPwGh6UUsA+CvndwDgBd0S3WgIdWvAvHM8wKgNh/GBLLzhT
Bg9BouRUzcT1MjAnkGkWqqCAgN7WrCSUMLt57TNleNWfX90CQQDjvVKTT3pOiavD
3YcLxwkyeGd0VMvKiS4nV0XXJ97cGXs2GpOGXldstDTnF5AnB6PbukdFLHpsx4sW
Hft3LRWnAkEA1pY15ke08wX6DZVXy7zuQ2izTrWSGySn7B41pn55dlKpttjHeutA
3BEQKTFvMhBCphr8qST7Wf1SR9FgO0tFbwJAEhHji2yy96hUyKW7IWQZhrem/cP8
p4Va9CQolnnDZRNgg1p4eiDiLu3dhLiJ547joXuWTBbLX/Y1Qvv+B+a74QJBAMCW
O3WbMZlS6eK6//rIa4ZwN00SxDg8I8FUM45jwBsjgVGrKQz2ilV3sutlhIiH82kk
m1Iq8LMJGYl/LkDJA10CQBV1C+Xu3ukknr7C4A/4lDCa6Xb27cr1HanY7i89A+Ab
eatdM6f/XVqWp8uPT9RggUV9TjppJobYGT2WrWJMkYw=
-----END RSA PRIVATE KEY-----
"""


class FakeHttpResponse:
    def __init__(self, status, headers, output):
        self.status = status
        self.__headers = headers
        self.__output = output

    def getheaders(self):
        return self.__headers

    def read(self):
        return self.__output

    def iter_content(self, chunk_size=1):
        return iter([self.__output[i : i + chunk_size] for i in range(0, len(self.__output), chunk_size)])

    def raise_for_status(self):
        pass


def fixAuthorizationHeader(headers):
    if "Authorization" in headers:
        if headers["Authorization"].endswith("ZmFrZV9sb2dpbjpmYWtlX3Bhc3N3b3Jk"):
            # This special case is here to test the real Authorization header
            # sent by PyGithub. It would have avoided issue https://github.com/jacquev6/PyGithub/issues/153
            # because we would have seen that Python 3 was not generating the same
            # header as Python 2
            pass
        elif headers["Authorization"].startswith("token "):
            headers["Authorization"] = "token private_token_removed"
        elif headers["Authorization"].startswith("Basic "):
            headers["Authorization"] = "Basic login_and_password_removed"
        elif headers["Authorization"].startswith("Bearer "):
            headers["Authorization"] = "Bearer jwt_removed"


@dataclass
class Request:
    protocol: str
    verb: str
    host: str
    port: int | None
    url: str
    request_headers: dict[str, Any]
    input: Any

    def with_response(self, status: int, response_headers: dict[str, Any], output: bytes) -> RequestResponse:
        return RequestResponse(
            self.protocol,
            self.verb,
            self.host,
            self.port,
            self.url,
            self.request_headers,
            self.input,
            status,
            response_headers,
            output,
        )


@dataclass
class RequestResponse(Request):
    protocol: str
    verb: str
    host: str
    port: int | None
    url: str
    request_headers: dict[str, Any]
    input: Any
    status: int
    response_headers: dict[str, Any]
    output: bytes


class Connection:
    __openFile: Callable[[str], ReplayDataFile] | None = None
    __requests: list[RequestResponse] = []

    @classmethod
    def setOpenFile(cls, func: Callable[[str], ReplayDataFile]):
        cls.__openFile = func

    @classmethod
    def openFile(cls, mode: str):
        assert cls.__openFile is not None
        return cls.__openFile(mode)

    @classmethod
    def resetRequests(cls, requests: list[RequestResponse]) -> list[RequestResponse]:
        try:
            return cls.__requests
        finally:
            cls.__requests = requests

    @classmethod
    def addRequest(cls, request: RequestResponse):
        cls.__requests.append(request)


class RecordingConnection(Connection):
    def __init__(self, protocol, host, port, *args, **kwds):
        self.__file = self.openFile("w")
        # write operations make the assumption that the file is not in binary mode
        assert isinstance(self.__file, ReplayDataFile)
        self.__request = None
        self.__protocol = protocol
        self.__host = host
        self.__port = port
        self.__cnx = self._realConnection(host, port, *args, **kwds)
        self.__stream = False

    @property
    def host(self):
        return self.__host

    def request(self, verb, url, input, headers, stream=False):
        self.__cnx.request(verb, url, input, headers)
        self.__stream = stream
        # fixAuthorizationHeader changes the parameter directly to remove Authorization token.
        # however, this is the real dictionary that *will be sent* by "requests",
        # since we are writing here *before* doing the actual request.
        # So we must avoid changing the real "headers" or this create this:
        # https://github.com/PyGithub/PyGithub/pull/664#issuecomment-389964369
        # https://github.com/PyGithub/PyGithub/issues/822
        # Since it's dict[str, str], a simple copy is enough.
        anonymous_headers = headers.copy()
        fixAuthorizationHeader(anonymous_headers)
        self.__request = Request(self.__protocol, verb, self.__host, self.__port, url, anonymous_headers, input)
        self.__writeLine(self.__protocol)
        self.__writeLine(verb)
        self.__writeLine(self.__host)
        self.__writeLine(self.__port)
        self.__writeLine(url)
        self.__writeLine(anonymous_headers)
        self.__writeLine(str(input).replace("\n", "").replace("\r", ""))

    def getresponse(self):
        res = self.__cnx.getresponse()

        status = res.status
        headers = res.getheaders()
        output = res if self.__stream else res.read()

        self.__writeLine(status)
        self.__writeLine(list(headers))
        if self.__stream:
            chunks = [chunk for chunk in output.iter_content(chunk_size=64)]
            output = b"".join(chunks)
            for chunk in chunks:
                self.__writeLine(base64.b64encode(chunk).decode("ascii"))
            self.__writeLine("")
        else:
            self.__writeLine(output)
        self.__writeLine("")

        self.addRequest(self.__request.with_response(status, dict(headers), output))
        self.__request = None

        return FakeHttpResponse(status, headers, output)

    def close(self):
        return self.__cnx.close()

    def __writeLine(self, line):
        self.__file.write(str(line) + "\n")


class RecordingHttpConnection(RecordingConnection):
    _realConnection = github.Requester.HTTPRequestsConnectionClass

    def __init__(self, *args, **kwds):
        super().__init__("http", *args, **kwds)


class RecordingHttpsConnection(RecordingConnection):
    _realConnection = github.Requester.HTTPSRequestsConnectionClass

    def __init__(self, *args, **kwds):
        super().__init__("https", *args, **kwds)


class ReplayingConnection(Connection):
    def __init__(self, protocol, host, port, *args, **kwds):
        self.__file = self.openFile("r")
        self.__protocol = protocol
        self.__host = host
        self.__port = port
        self.__stream = False
        self.response_headers = CaseInsensitiveDict()

        self.__cnx = self._realConnection(host, port, *args, **kwds)

    @property
    def host(self):
        return self.__host

    def request(
        self,
        verb,
        url,
        input,
        headers,
        stream: bool = False,
    ):
        port = self.__port if self.__port else 443 if self.__protocol == "https" else 80
        full_url = Url(scheme=self.__protocol, host=self.__host, port=port, path=url)

        response_headers = self.response_headers.copy()
        responses.add_callback(
            method=verb,
            url=full_url.url,
            callback=lambda request: self.__request_callback(verb, full_url.url, response_headers),
        )

        self.__stream = stream
        self.__cnx.request(verb, url, input, headers, stream=stream)

    def __replayDataMismatchLine(self) -> str:
        return f"Replay data mismatch in {self.__file}"

    def __readNextRequest(self, verb, url, input, headers) -> Request:
        fixAuthorizationHeader(headers)
        request = Request(self.__protocol, verb, self.__host, self.__port, url, headers, input)
        assert request.protocol == self.__file.readline(), self.__replayDataMismatchLine()
        assert request.verb == self.__file.readline(), self.__replayDataMismatchLine()
        assert request.host == self.__file.readline(), self.__replayDataMismatchLine()
        assert str(request.port) == self.__file.readline(), self.__replayDataMismatchLine()
        assert self.__splitUrl(request.url) == self.__splitUrl(self.__file.readline()), self.__replayDataMismatchLine()
        assert request.request_headers == eval(self.__file.readline()), self.__replayDataMismatchLine()
        expectedInput = self.__file.readline()
        if isinstance(input, str):
            trInput = input.replace("\n", "").replace("\r", "")
            if input.startswith("{"):
                assert expectedInput.startswith("{"), self.__replayDataMismatchLine()
                assert json.loads(trInput) == json.loads(expectedInput), self.__replayDataMismatchLine()
            else:
                assert trInput == expectedInput, self.__replayDataMismatchLine()
        else:
            # for non-string input (e.g. upload asset), let it pass.
            pass

        return request

    def __splitUrl(self, url):
        splitedUrl = url.split("?")
        if len(splitedUrl) == 1:
            return splitedUrl
        assert len(splitedUrl) == 2
        base, qs = splitedUrl
        return (base, sorted(qs.split("&")))

    def __request_callback(self, request, uri, response_headers):
        request = self.__readNextRequest(self.__cnx.verb, self.__cnx.url, self.__cnx.input, self.__cnx.headers)

        status = int(self.__file.readline())
        self.response_headers = CaseInsensitiveDict(eval(self.__file.readline()))
        if self.__stream:
            output = BytesIO()
            while True:
                line = self.__file.readline()
                if not line:
                    break
                output.write(base64.b64decode(line))
            output = output.getvalue()
        else:
            output = bytearray(self.__file.readline(), "utf-8")
        self.__file.readline()

        # make a copy of the headers and remove the ones that interfere with the response handling
        adding_headers = CaseInsensitiveDict(self.response_headers)
        adding_headers.pop("content-length", None)
        adding_headers.pop("transfer-encoding", None)
        adding_headers.pop("content-encoding", None)

        response_headers.update(adding_headers)
        self.addRequest(request.with_response(status, response_headers, output))

        return [status, response_headers, output]

    def getresponse(self):
        response = self.__cnx.getresponse()

        # restore original headers to the response
        response.headers = self.response_headers

        return response

    def close(self):
        self.__cnx.close()


class ReplayingHttpConnection(ReplayingConnection):
    _realConnection = github.Requester.HTTPRequestsConnectionClass

    def __init__(self, *args, **kwds):
        super().__init__("http", *args, **kwds)


class ReplayingHttpsConnection(ReplayingConnection):
    _realConnection = github.Requester.HTTPSRequestsConnectionClass

    def __init__(self, *args, **kwds):
        super().__init__("https", *args, **kwds)


class ReplayDataFile:
    @staticmethod
    def open(filename: str, mode: str, encoding: str) -> ReplayDataFile:
        file = open(filename, mode, encoding=encoding)
        return ReplayDataFile(filename, file)

    def __init__(self, filename: str, file):
        self.__filename = filename
        self.__file = file
        self.__line = 0

    def __repr__(self) -> str:
        return f"{self.__filename}:{self.__line}"

    def write(self, string: str):
        self.__file.write(string)

    def readline(self) -> str:
        self.__line += 1
        line = self.__file.readline()
        if isinstance(line, bytes):
            line = line.decode("utf-8")
        return line.strip()

    @property
    def line_number(self):
        return self.__line

    def close(self):
        self.__file.close()


class BasicTestCase(unittest.TestCase):
    recordMode = False
    replayDataFolder = os.path.join(os.path.dirname(__file__), "ReplayData")

    def __init__(self, methodName="runTest") -> None:
        super().__init__(methodName)
        self.authMode = "token"
        self.per_page = Consts.DEFAULT_PER_PAGE
        self.retry = None
        self.pool_size = None
        self.seconds_between_requests: float | None = None
        self.seconds_between_writes: float | None = None

    def setUp(self):
        super().setUp()
        self.__customFilename: str | None = None
        self.__fileName = ""
        self.__file = None
        if (
            self.recordMode
        ):  # pragma no cover (Branch useful only when recording new tests, not used during automated tests)
            RecordingHttpConnection.setOpenFile(self.__openFile)
            RecordingHttpsConnection.setOpenFile(self.__openFile)
            github.Requester.Requester.injectConnectionClasses(
                RecordingHttpConnection,
                RecordingHttpsConnection,
            )
            import GithubCredentials  # type: ignore

            self.oauth_token = (
                github.Auth.Token(GithubCredentials.oauth_token) if GithubCredentials.oauth_token else None
            )
            self.jwt = github.Auth.AppAuthToken(GithubCredentials.jwt) if GithubCredentials.jwt else None
            self.app_auth = (
                github.Auth.AppAuth(GithubCredentials.app_id, GithubCredentials.app_private_key)
                if GithubCredentials.app_id and GithubCredentials.app_private_key
                else None
            )
        else:
            ReplayingHttpConnection.setOpenFile(self.__openFile)
            ReplayingHttpsConnection.setOpenFile(self.__openFile)
            github.Requester.Requester.injectConnectionClasses(
                ReplayingHttpConnection,
                ReplayingHttpsConnection,
            )
            self.oauth_token = github.Auth.Token("oauth_token")
            self.jwt = github.Auth.AppAuthToken("jwt")
            self.app_auth = github.Auth.AppAuth(123456, APP_PRIVATE_KEY)

            responses.start()

    def setPerPage(self, per_page):
        self.per_page = per_page

    def setRetry(self, retry):
        self.retry = retry

    def setPoolSize(self, pool_size):
        self.pool_size = pool_size

    def setSecondsBetweenRequests(self, seconds_between_requests):
        self.seconds_between_requests = seconds_between_requests

    def setSecondsBetweenWrites(self, seconds_between_writes):
        self.seconds_between_writes = seconds_between_writes

    @property
    def thisTestFailed(self) -> bool:
        if hasattr(self._outcome, "errors"):  # type: ignore
            # Python 3.4 - 3.10
            result = self.defaultTestResult()
            self._feedErrorsToResult(result, self._outcome.errors)  # type: ignore
            ok = all(test != self for test, text in result.errors + result.failures)
            return not ok
        else:
            # Python 3.11+
            return self._outcome.result._excinfo is not None and self._outcome.result._excinfo  # type: ignore

    def tearDown(self):
        super().tearDown()
        responses.stop()
        responses.reset()
        self.__closeReplayFileIfNeeded(silent=self.thisTestFailed)
        github.Requester.Requester.resetConnectionClasses()

    def assertWarning(self, warning, expected):
        self.assertWarnings(warning, expected)

    def assertWarnings(self, warning, *expecteds):
        actual = [(type(message), type(message.message), message.message.args) for message in warning.warnings]
        expected = [(warnings.WarningMessage, DeprecationWarning, (expected,)) for expected in expecteds]
        self.assertSequenceEqual(actual, expected)

    @contextlib.contextmanager
    def ignoreWarning(self, category=Warning, module=""):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=category, module=module)
            yield

    @contextlib.contextmanager
    def replayData(self, filename: str):
        previous = self.__customFilename
        self.__customFilename = filename
        try:
            yield
        finally:
            self.__customFilename = previous

    @contextlib.contextmanager
    def captureRequests(self) -> Generator[list[RequestResponse]]:
        requests: list[RequestResponse] = []
        earlier_requests = Connection.resetRequests(requests)
        try:
            yield requests
        finally:
            earlier_requests.extend(requests)
            Connection.resetRequests(earlier_requests)

    def __openFile(self, mode):
        fileName = None
        if self.__customFilename:
            fileName = self.__customFilename
        else:
            for _, _, functionName, _ in traceback.extract_stack():
                if functionName.startswith("test") or functionName == "setUp" or functionName == "tearDown":
                    # because in class Hook(Framework.TestCase), method testTest calls Hook.test
                    if functionName != "test":
                        fileName = f"{self.__class__.__name__}.{functionName}.txt"
        fileName = os.path.join(self.replayDataFolder, fileName) if fileName else None
        if fileName != self.__fileName:
            self.__closeReplayFileIfNeeded()
            self.__fileName = fileName
            self.__file = ReplayDataFile.open(self.__fileName, mode, encoding="utf-8")
        return self.__file

    def __closeReplayFileIfNeeded(self, silent=False):
        if self.__file is not None:
            if (
                not self.recordMode and not silent
            ):  # pragma no branch (Branch useful only when recording new tests, not used during automated tests)
                self.assertEqual(self.__file.readline(), "", self.__file)
            self.__file.close()

    def assertListKeyEqual(self, elements, key, expectedKeys):
        realKeys = [key(element) for element in elements]
        self.assertEqual(realKeys, expectedKeys)

    def assertListKeyBegin(self, elements, key, expectedKeys):
        realKeys = [key(element) for element in elements[: len(expectedKeys)]]
        self.assertEqual(realKeys, expectedKeys)


class TestCase(BasicTestCase):
    def doCheckFrame(self, obj, frame):
        if obj._headers == {} and frame is None:
            return
        if obj._headers is None and frame == {}:
            return
        self.assertEqual(obj._headers, frame[2])

    def getFrameChecker(self):
        return lambda requester, obj, frame: self.doCheckFrame(obj, frame)

    def setUp(self):
        super().setUp()

        # Set up frame debugging
        github.GithubObject.GithubObject.setCheckAfterInitFlag(True)
        github.Requester.Requester.setDebugFlag(True)
        github.Requester.Requester.setOnCheckMe(self.getFrameChecker())

        self.g = self.get_github(self.authMode, self.retry, self.pool_size)

    def get_github(self, authMode, retry=None, pool_size=None):
        if authMode == "token":
            auth = self.oauth_token
        elif authMode == "jwt":
            auth = self.jwt
        elif authMode == "app":
            auth = self.app_auth
        elif self.authMode == "none":
            auth = None
        else:
            raise ValueError(f"Unsupported test auth mode: {authMode}")

        return github.Github(
            auth=auth,
            per_page=self.per_page,
            retry=retry,
            pool_size=pool_size,
            seconds_between_requests=self.seconds_between_requests,
            seconds_between_writes=self.seconds_between_writes,
        )


def activateRecordMode():  # pragma no cover (Function useful only when recording new tests, not used during automated tests)
    BasicTestCase.recordMode = True



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2023 Andrew Dawes <53574062+AndrewJDawes@users.noreply.github.com> #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2023 alson <git@alm.nufan.net>                                     #
# Copyright 2024 Kian-Meng Ang <kianmeng.ang@gmail.com>                        #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2025 GPK <gopidesupavan@gmail.com>                                 #
# Copyright 2025 Oscar van Leusen <oscarvanleusen@gmail.com>                   #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from datetime import datetime, timezone
from unittest import mock

import pytest  # type: ignore

import github
import github.EnvironmentDeploymentBranchPolicy
import github.EnvironmentProtectionRule
import github.EnvironmentProtectionRuleReviewer
import github.NamedUser
import github.Team

from . import Framework


class Environment(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.repo = self.g.get_user().get_repo("PyGithub")
        self.environment = self.repo.get_environment("dev")

    def testAttributes(self):
        self.assertEqual(self.environment.name, "dev")
        self.assertEqual(self.environment.id, 464814513)
        self.assertEqual(self.environment.node_id, "EN_kwDOHKhL9c4btIGx")
        self.assertListKeyEqual(self.environment.protection_rules, lambda r: r.id, [216323, 216324, 216325])
        self.assertEqual(
            self.environment.url,
            "https://api.github.com/repos/alson/PyGithub/environments/dev",
        )
        self.assertEqual(
            self.environment.html_url,
            "https://github.com/alson/PyGithub/deployments/activity_log?environments_filter=dev",
        )
        self.assertEqual(
            self.environment.created_at,
            datetime(2022, 4, 13, 15, 6, 32, tzinfo=timezone.utc),
        )
        self.assertEqual(
            self.environment.updated_at,
            datetime(2022, 4, 13, 15, 6, 32, tzinfo=timezone.utc),
        )
        self.assertTrue(self.environment.deployment_branch_policy.protected_branches)
        self.assertFalse(self.environment.deployment_branch_policy.custom_branch_policies)

    def testLazyAttributes(self):
        env = self.g.withLazy(True).get_repo("lazy/repo").get_environment("the env")
        self.assertEqual(str(env), 'Environment(name="the env")')
        self.assertEqual(env.name, "the env")
        self.assertEqual(env.url, "/repos/lazy/repo/environments/the%20env")

    def testProtectionRules(self):
        protection_rules = self.environment.protection_rules
        self.assertEqual(len(protection_rules), 3)
        self.assertEqual(protection_rules[0].id, 216323)
        self.assertEqual(protection_rules[0].node_id, "GA_kwDOHKhL9c4AA00D")
        self.assertEqual(protection_rules[0].type, "branch_policy")
        self.assertEqual(protection_rules[1].id, 216324)
        self.assertEqual(protection_rules[1].node_id, "GA_kwDOHKhL9c4AA00E")
        self.assertEqual(protection_rules[1].type, "required_reviewers")
        self.assertFalse(protection_rules[1].prevent_self_review)
        self.assertEqual(protection_rules[2].id, 216325)
        self.assertEqual(protection_rules[2].node_id, "GA_kwDOHKhL9c4AA00F")
        self.assertEqual(protection_rules[2].type, "wait_timer")
        self.assertEqual(protection_rules[2].wait_timer, 15)

    def testReviewers(self):
        # This is necessary so we can maintain our own expectations, which have been manually edited, for this test.
        reviewers = self.repo.get_environment("dev").protection_rules[1].reviewers
        self.assertEqual(len(reviewers), 2)
        self.assertEqual(reviewers[0].type, "User")
        self.assertIsInstance(reviewers[0].reviewer, github.NamedUser.NamedUser)
        # Make type checker happy
        assert isinstance(reviewers[0].reviewer, github.NamedUser.NamedUser)
        self.assertEqual(reviewers[0].reviewer.id, 19245)
        self.assertEqual(reviewers[0].reviewer.login, "alson")
        self.assertEqual(reviewers[0].reviewer.type, "User")
        self.assertEqual(reviewers[1].type, "Team")
        self.assertIsInstance(reviewers[1].reviewer, github.Team.Team)
        # Make type checker happy
        assert isinstance(reviewers[1].reviewer, github.Team.Team)
        self.assertEqual(reviewers[1].reviewer.id, 1)
        self.assertEqual(reviewers[1].reviewer.slug, "justice-league")
        self.assertEqual(reviewers[1].reviewer.url, "https://api.github.com/teams/1")

    def testGetEnvironments(self):
        environments = self.repo.get_environments()
        self.assertEqual(environments.totalCount, 1)
        self.assertEqual(
            environments[0].url,
            "https://api.github.com/repos/alson/PyGithub/environments/dev",
        )
        self.assertEqual(environments[0].name, "dev")

    def testCreateEnvironment(self):
        environment = self.repo.create_environment("test/env")
        self.assertEqual(environment.name, "test/env")
        self.assertEqual(environment.id, 470015651)
        self.assertEqual(environment.node_id, "EN_kwDOHKhL9c4cA96j")
        self.assertEqual(
            environment.url,
            "https://api.github.com/repos/alson/PyGithub/environments/test/env",
        )
        self.assertEqual(
            environment.html_url,
            "https://github.com/alson/PyGithub/deployments/activity_log?environments_filter=test%2Fenv",
        )
        self.assertEqual(
            environment.created_at,
            datetime(2022, 4, 19, 14, 4, 32, tzinfo=timezone.utc),
        )
        self.assertEqual(
            environment.updated_at,
            datetime(2022, 4, 19, 14, 4, 32, tzinfo=timezone.utc),
        )
        self.assertEqual(len(environment.protection_rules), 0)
        self.assertIsNone(environment.deployment_branch_policy)

    def testUpdateEnvironment(self):
        environment = self.repo.create_environment(
            "test/env",
            wait_timer=42,
            reviewers=[github.EnvironmentProtectionRuleReviewer.ReviewerParams(type_="User", id_=19245)],
            prevent_self_review=True,
            deployment_branch_policy=github.EnvironmentDeploymentBranchPolicy.EnvironmentDeploymentBranchPolicyParams(
                protected_branches=True, custom_branch_policies=False
            ),
        )
        self.assertEqual(environment.name, "test")
        self.assertEqual(environment.id, 470015651)
        self.assertEqual(environment.node_id, "EN_kwDOHKhL9c4cA96j")
        self.assertEqual(
            environment.url,
            "https://api.github.com/repos/alson/PyGithub/environments/test/env",
        )
        self.assertEqual(
            environment.html_url,
            "https://github.com/alson/PyGithub/deployments/activity_log?environments_filter=test%2Fenv",
        )
        self.assertEqual(
            environment.created_at,
            datetime(2022, 4, 19, 14, 4, 32, tzinfo=timezone.utc),
        )
        self.assertEqual(
            environment.updated_at,
            datetime(2022, 4, 19, 14, 4, 32, tzinfo=timezone.utc),
        )
        self.assertEqual(len(environment.protection_rules), 3)
        self.assertEqual(environment.protection_rules[0].type, "required_reviewers")
        self.assertTrue(environment.protection_rules[0].prevent_self_review)
        self.assertEqual(len(environment.protection_rules[0].reviewers), 1)
        self.assertEqual(environment.protection_rules[0].reviewers[0].type, "User")
        self.assertEqual(environment.protection_rules[0].reviewers[0].reviewer.id, 19245)
        self.assertEqual(environment.protection_rules[1].type, "wait_timer")
        self.assertEqual(environment.protection_rules[1].wait_timer, 42)
        self.assertEqual(environment.protection_rules[2].type, "branch_policy")
        self.assertTrue(environment.deployment_branch_policy.protected_branches)
        self.assertFalse(environment.deployment_branch_policy.custom_branch_policies)

    def testDeleteEnvironment(self):
        self.repo.delete_environment("test")
        with pytest.raises(github.UnknownObjectException):
            self.repo.get_environment("test")

    def testEnvironmentVariable(self):
        repo = self.g.get_repo("AndrewJDawes/PyGithub")
        environment = repo.create_environment("test")
        variable = environment.create_variable("variable_name", "variable-value")
        self.assertTrue(variable.edit("variable-value123"))
        variable.delete()
        repo.delete_environment("test")

    def testEnvironmentVariables(self):
        # GitHub will always capitalize the variable name
        variables = (("VARIABLE_NAME_ONE", "variable-value-one"), ("VARIABLE_NAME_TWO", "variable-value-two"))
        repo = self.g.get_repo("AndrewJDawes/PyGithub")
        environment = repo.create_environment("test")
        for variable in variables:
            environment.create_variable(variable[0], variable[1])
        environment.update()
        environment_variables = environment.get_variables()
        matched_environment_variables = []
        for variable in variables:
            for environment_variable in environment_variables:
                # GitHub will always capitalize the variable name, may be best to uppercase test data for comparison
                if environment_variable.name == variable[0].upper() and environment_variable.value == variable[1]:
                    matched_environment_variables.append(environment_variable)
                    break
        self.assertEqual(len(matched_environment_variables), len(variables))
        for matched_environment_variable in matched_environment_variables:
            matched_environment_variable.delete()
        repo.delete_environment("test")

    @mock.patch("github.PublicKey.encrypt")
    def testEnvironmentSecret(self, encrypt):
        # encrypt returns a non-deterministic value, we need to mock it so the replay data matches
        encrypt.return_value = "M+5Fm/BqTfB90h3nC7F3BoZuu3nXs+/KtpXwxm9gG211tbRo0F5UiN0OIfYT83CKcx9oKES9Va4E96/b"
        repo = self.g.get_repo("AndrewJDawes/PyGithub")
        environment = repo.create_environment("test")
        secret = environment.create_secret("secret_name", "secret-value")
        secret.update()
        # GitHub will always capitalize the secret name
        self.assertEqual(secret.name, "SECRET_NAME")
        secret.delete()
        repo.delete_environment("test")

    @mock.patch("github.PublicKey.encrypt")
    def testEnvironmentSecrets(self, encrypt):
        # encrypt returns a non-deterministic value, we need to mock it so the replay data matches
        encrypt.return_value = "M+5Fm/BqTfB90h3nC7F3BoZuu3nXs+/KtpXwxm9gG211tbRo0F5UiN0OIfYT83CKcx9oKES9Va4E96/b"
        # GitHub will always capitalize the secret name
        secrets = (("SECRET_NAME_ONE", "secret-value-one"), ("SECRET_NAME_TWO", "secret-value-two"))
        repo = self.g.get_repo("AndrewJDawes/PyGithub")
        environment = repo.create_environment("test")
        for secret in secrets:
            environment.create_secret(secret[0], secret[1])
        environment.update()
        environment_secrets = environment.get_secrets()
        matched_environment_secrets = []
        for secret in secrets:
            for environment_secret in environment_secrets:
                # GitHub will always capitalize the secret name, may be best to uppercase test data for comparison
                if environment_secret.name == secret[0].upper():
                    matched_environment_secrets.append(environment_secret)
                    break
        self.assertEqual(len(matched_environment_secrets), len(secrets))
        for matched_environment_secret in matched_environment_secrets:
            matched_environment_secret.delete()
        repo.delete_environment("test")



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Jannis Gebauer <ja.geb@me.com>                                #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Pascal Hofmann <mail@pascalhofmann.de>                        #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2023 Nevins <nevins-b@users.noreply.github.com>                    #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from datetime import datetime, timezone

from . import Framework


class Deployment(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.deployment = self.g.get_user().get_repo("PyGithub").get_deployment(263877258)

    def testAttributes(self):
        self.assertEqual(self.deployment.created_at, datetime(2020, 8, 26, 11, 44, 53, tzinfo=timezone.utc))
        self.assertEqual(self.deployment.creator.login, "jacquev6")
        self.assertEqual(self.deployment.description, "Test deployment")
        self.assertEqual(self.deployment.environment, "test")
        self.assertEqual(self.deployment.id, 263877258)
        self.assertEqual(self.deployment.node_id, "MDEwOkRlcGxveW1lbnQyNjIzNTE3NzY=")
        self.assertEqual(self.deployment.original_environment, "test")
        self.assertEqual(self.deployment.payload, {"test": True})
        self.assertIsNone(self.deployment.performed_via_github_app)
        self.assertEqual(self.deployment.production_environment, False)
        self.assertEqual(self.deployment.ref, "743f5a58b0bce91c4eab744ff7e39dfca9e6e8a5")
        self.assertEqual(self.deployment.repository_url, "https://api.github.com/repos/jacquev6/PyGithub")
        self.assertEqual(self.deployment.sha, "743f5a58b0bce91c4eab744ff7e39dfca9e6e8a5")
        self.assertEqual(
            self.deployment.statuses_url,
            "https://api.github.com/repos/jacquev6/PyGithub/deployments/263877258/statuses",
        )
        self.assertEqual(self.deployment.task, "deploy")
        self.assertEqual(self.deployment.transient_environment, True)
        self.assertEqual(self.deployment.updated_at, datetime(2020, 8, 26, 11, 44, 53, tzinfo=timezone.utc))
        self.assertEqual(
            self.deployment.url,
            "https://api.github.com/repos/jacquev6/PyGithub/deployments/263877258",
        )
        self.assertEqual(self.deployment.ref, "743f5a58b0bce91c4eab744ff7e39dfca9e6e8a5")
        self.assertEqual(self.deployment.sha, "743f5a58b0bce91c4eab744ff7e39dfca9e6e8a5")
        self.assertEqual(self.deployment.task, "deploy")
        self.assertEqual(self.deployment.payload, {"test": True})
        self.assertEqual(self.deployment.original_environment, "test")
        self.assertEqual(self.deployment.environment, "test")
        self.assertEqual(self.deployment.description, "Test deployment")
        self.assertEqual(self.deployment.creator.login, "jacquev6")
        created_at = datetime(2020, 8, 26, 11, 44, 53, tzinfo=timezone.utc)
        self.assertEqual(self.deployment.created_at, created_at)
        self.assertEqual(self.deployment.updated_at, created_at)
        self.assertEqual(self.deployment.transient_environment, True)
        self.assertEqual(self.deployment.production_environment, False)
        self.assertEqual(
            self.deployment.statuses_url,
            "https://api.github.com/repos/jacquev6/PyGithub/deployments/263877258/statuses",
        )
        self.assertEqual(
            self.deployment.repository_url,
            "https://api.github.com/repos/jacquev6/PyGithub",
        )
        self.assertEqual(
            repr(self.deployment),
            'Deployment(url="https://api.github.com/repos/jacquev6/PyGithub/deployments/263877258", id=263877258)',
        )

    def testLazyAttributes(self):
        deploy = self.g.withLazy(True).get_repo("lazy/repo").get_deployment(42)
        self.assertEqual(str(deploy), 'Deployment(url="/repos/lazy/repo/deployments/42", id=42)')
        self.assertEqual(deploy.id, 42)
        self.assertEqual(deploy.url, "/repos/lazy/repo/deployments/42")



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2025 Hugo van Kemenade <1324225+hugovk@users.noreply.github.com>   #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

import argparse
import difflib
import json
import multiprocessing
import sys
from collections.abc import Sequence

import libcst as cst


class SortMethodsTransformer(cst.CSTTransformer):
    def __init__(self, class_name: str | None = None, sort_funcs: bool = False):
        super().__init__()
        self.visit_class_name = []
        self.class_name = class_name
        self.sort_funcs = sort_funcs

    @property
    def current_class_name(self) -> str:
        return ".".join(self.visit_class_name)

    def visit_ClassDef(self, node: cst.ClassDef):
        self.visit_class_name.append(node.name.value)

    def leave_ClassDef(self, original_node: cst.ClassDef, updated_node: cst.ClassDef):
        try:
            if self.class_name is not None and self.current_class_name != self.class_name:
                return updated_node
            if not any(
                base.value.value.endswith("GithubObject")
                if isinstance(base.value, cst.Name)
                else (
                    base.value.value.value.endswith("GithubObject")
                    if isinstance(base.value.value, cst.Name)
                    else base.value.value.attr.value.endswith("GithubObject")
                )
                for base in updated_node.bases
            ):
                return updated_node

            statements = list(updated_node.body.body)
            if sum(1 for s in statements if isinstance(s, cst.FunctionDef)) == 0:
                raise ValueError(f"There are no functions in class {self.current_class_name}")
            first = next(idx for idx, s in enumerate(statements) if isinstance(s, cst.FunctionDef))
            last = len(statements) - next(
                idx for idx, s in enumerate(reversed(statements)) if isinstance(s, cst.FunctionDef)
            )
            if any(not isinstance(s, cst.FunctionDef) for s in statements[first:last]):
                raise ValueError(f"There is no consecutive block of functions in class {self.current_class_name}")

            # noinspection PyTypeChecker
            function_defs: list[cst.FunctionDef] = statements[first:last]
            prolog = statements[:first]

            init = self.find_func(function_defs, "__init__")
            init = [init] if init is not None else []

            init_attrs = self.find_func(function_defs, "_initAttributes")
            init_attrs = [init_attrs] if init_attrs is not None else []

            dunders = list(
                s
                for s in function_defs
                if s.name.value.startswith("__") and s.name.value.endswith("__") and s.name.value != "__init__"
            )
            properties = list(s for s in function_defs if self.contains_decorator(s.decorators, "property"))

            use_attrs = self.find_func(function_defs, "_useAttributes")
            use_attrs = [use_attrs] if use_attrs is not None else []

            funcs = list(s for s in function_defs if s not in init + init_attrs + use_attrs + dunders + properties)
            epilog = statements[last:]

            sorted_dunders = self.sort_func_defs(dunders)
            sorted_properties = self.sort_func_defs(properties)
            maybe_sorted_funcs = self.sort_func_defs(funcs) if self.sort_funcs else funcs
            sorted_functions = (
                prolog
                + init
                + init_attrs
                + sorted_dunders
                + sorted_properties
                + maybe_sorted_funcs
                + use_attrs
                + epilog
            )
            return updated_node.with_changes(body=updated_node.body.with_changes(body=sorted_functions))

        finally:
            self.visit_class_name.pop()

    def find_func(self, funcs: Sequence[cst.FunctionDef], func_name: str) -> cst.FunctionDef | None:
        funcs = list(s for s in funcs if s.name.value == func_name)
        if len(funcs) == 0:
            return None
        if len(funcs) > 1:
            raise ValueError(f"Multiple functions {func_name} exist in class {self.current_class_name}")
        return funcs[0]

    @staticmethod
    def contains_decorator(seq: Sequence[cst.Decorator], decorator_name: str):
        return any(d.decorator.value == decorator_name for d in seq if isinstance(d.decorator, cst.Name))

    @staticmethod
    def sort_func_defs(funcs: list[cst.FunctionDef]) -> list[cst.FunctionDef]:
        return sorted(funcs, key=lambda d: d.name.value)

    def leave_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef):
        if self.class_name is not None and self.current_class_name != self.class_name:
            return updated_node
        if updated_node.name.value == "_initAttributes":
            attrs = list(
                [
                    idx
                    for idx, stmt in enumerate(updated_node.body.body)
                    if isinstance(stmt, cst.SimpleStatementLine) and isinstance(stmt.body[0], cst.AnnAssign)
                ]
            )
            if attrs:
                start_attr, end_attr = attrs[0], attrs[-1] + 1
                stmts = updated_node.body.body
                attrs = sorted(
                    stmts[start_attr:end_attr],
                    key=lambda stmt: stmt.body[0].target.attr.value
                    if isinstance(stmt.body[0], cst.AnnAssign)
                    else stmt.body[0].targets[0].target.attr.value,
                )
                updated_node = updated_node.with_changes(
                    body=updated_node.body.with_changes(
                        body=tuple(stmts[:start_attr]) + tuple(attrs) + tuple(stmts[end_attr:])
                    )
                )
        if updated_node.name.value == "_useAttributes":
            ifs = list([idx for idx, stmt in enumerate(updated_node.body.body) if isinstance(stmt, cst.If)])
            if ifs:
                start_if, end_if = ifs[0], ifs[-1] + 1
                stmts = updated_node.body.body
                ifs = sorted(
                    stmts[start_if:end_if],
                    key=lambda stmt: stmt.test.left.value
                    if isinstance(stmt.test, cst.Comparison)
                    else stmt.test.children[0].left.value,
                )
                updated_node = updated_node.with_changes(
                    body=updated_node.body.with_changes(
                        body=tuple(stmts[:start_if]) + tuple(ifs) + tuple(stmts[end_if:])
                    )
                )
        return updated_node


def sort_class(class_name: str, filename: str, dry_run: bool, locks: dict[str, multiprocessing.Lock]):
    print(f"Sorting {class_name} ({filename})")

    file_lock = locks.get(filename) if not dry_run else None
    stdout_lock = locks.get("stdout")

    if file_lock:
        file_lock.acquire()
    try:
        with open(filename) as r:
            code = "".join(r.readlines())

        tree = cst.parse_module(code)
        transformer = SortMethodsTransformer(class_name)
        tree_updated = tree.visit(transformer)

        if dry_run:
            diff = "".join(difflib.unified_diff(code.splitlines(1), tree_updated.code.splitlines(1)))
            if diff:
                stdout_lock.acquire()
                print(f"Diff of {class_name}:")
                print(diff)
                print()
                stdout_lock.release()
        else:
            if not tree_updated.deep_equals(tree):
                with open(filename, "w") as w:
                    w.write(tree_updated.code)
    finally:
        if file_lock:
            file_lock.release()


class SortFileWorker:
    def __init__(self, dry_run: bool, locks: dict[str, multiprocessing.Lock]):
        self.dry_run = dry_run
        self.locks = locks

    def sort(self, class_name_and_file: tuple[str, str]):
        class_name, filename = class_name_and_file
        sort_class(class_name, filename, self.dry_run, self.locks)


def main(index_filename: str, class_names: list[str], dry_run: bool):
    if len(class_names) > 1:
        print(f"Sorting {len(class_names)} Python files")

    with open(index_filename) as r:
        index = json.load(r)

    filenames = {}
    for class_name in class_names:
        full_class_name = class_name
        if "." not in class_name:
            cls = index.get("classes", {}).get(class_name)
            if not cls:
                raise ValueError(f"Class {class_name} does not exist in index")
            full_class_name = f'{cls.get("package")}.{cls.get("module")}.{cls.get("name")}'
        package, module, class_name = full_class_name.split(".", maxsplit=2)
        filename = f"{package}/{module}.py"
        filenames[class_name] = filename

    with multiprocessing.Manager() as manager:
        locks = {filename: manager.Lock() for filename in filenames}
        locks.update({"stdout": manager.Lock()})

        # sort files in parallel
        worker = SortFileWorker(dry_run, locks)
        with multiprocessing.Pool() as pool:
            pool.map(worker.sort, iterable=filenames.items())


def parse_args():
    args_parser = argparse.ArgumentParser(
        description="Sorts methods of GithubObject classes, also sorts attributes in _initAttributes and _useAttributes"
    )
    args_parser.add_argument("index_filename", help="Path of index file")
    args_parser.add_argument(
        "class_name",
        nargs="+",
        help="GithubObject class to sort, e.g. HookDelivery or github.HookDelivery.HookDeliverySummary",
    )
    args_parser.add_argument(
        "--dry-run", default=False, action="store_true", help="show prospect changes and do not modify the file"
    )
    if len(sys.argv) == 1:
        args_parser.print_help()
        sys.exit(1)
    return args_parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    main(args.index_filename, args.class_name, args.dry_run)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2020 Victor Zeng <zacker150@users.noreply.github.com>              #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Trim21 <trim21.me@gmail.com>                                  #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from . import Framework


class SelfHostedActionsRunner(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.user = self.g.get_user("ReDASers")
        self.repo = self.user.get_repo("Phishing-Detection")

    def testAttributes(self):
        runner = self.repo.get_self_hosted_runner(2217)
        self.assertFalse(runner.busy)
        self.assertIsNone(runner.ephemeral)
        self.assertEqual(runner.id, 2217)
        self.assertEqual(
            runner.labels,
            [
                {"id": 1, "name": "self-hosted", "type": "read-only"},
                {"id": 3, "name": "X64", "type": "read-only"},
                {"id": 4, "name": "Linux", "type": "read-only"},
            ],
        )
        self.assertEqual(runner.name, "4306125c7c84")
        self.assertEqual(runner.os, "linux")
        self.assertEqual(runner.name, "4306125c7c84")
        self.assertIsNone(runner.runner_group_id)
        self.assertEqual(runner.status, "offline")
        labels = runner.labels
        self.assertEqual(len(labels), 3)
        self.assertEqual(labels[0]["name"], "self-hosted")
        self.assertEqual(labels[1]["name"], "X64")
        self.assertEqual(labels[2]["name"], "Linux")



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2017 Chris McBride <thehighlander@users.noreply.github.com>        #
# Copyright 2017 Simon <spam@esemi.ru>                                         #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2025 Alex Olieman <alex@olieman.net>                               #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import hashlib
import tempfile
from datetime import datetime, timezone

from . import Framework


class ReleaseAsset(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.release = self.g.get_user().get_repo("PyGithub").get_releases()[0]
        self.asset = self.release.get_assets()[0]

    def testAttributes(self):
        self.assertEqual(self.release.id, 1210814)
        self.assertEqual(self.asset.id, 16)
        self.assertEqual(
            self.asset.url,
            "https://api.github.com/api/v3/repos/edhollandAL/PyGithub/releases/assets/16",
        )
        self.assertEqual(self.asset.name, "Archive.zip")
        self.assertEqual(self.asset.label, "Installation msi & runbook zipped")
        self.assertEqual(self.asset.content_type, "application/zip")
        self.assertEqual(self.asset.state, "uploaded")
        self.assertEqual(self.asset.size, 3783)
        self.assertEqual(self.asset.download_count, 2)
        self.assertEqual(
            self.asset.created_at,
            datetime(2017, 2, 1, 22, 40, 58, tzinfo=timezone.utc),
        )
        self.assertEqual(
            self.asset.updated_at,
            datetime(2017, 2, 1, 22, 44, 58, tzinfo=timezone.utc),
        )
        self.assertEqual(
            self.asset.browser_download_url,
            "https://github.com/edhollandAL/PyGithub/releases/download/v1.25.2/Asset.zip",
        )
        self.assertEqual(self.asset.uploader.login, "PyGithub")
        self.assertEqual(
            repr(self.asset),
            'GitReleaseAsset(url="https://api.github.com/api/v3/repos/edhollandAL/PyGithub/releases/assets/16")',
        )

    def testDelete(self):
        self.assertTrue(self.asset.delete_asset())

    def testUpdate(self):
        new_name = "updated-name.zip"
        new_label = "Updated label"
        updated_asset = self.asset.update_asset(new_name, new_label)
        self.assertEqual(updated_asset.name, new_name)
        self.assertNotEqual(self.asset.name, updated_asset.name)
        self.assertEqual(updated_asset.label, new_label)
        self.assertNotEqual(self.asset.label, updated_asset.label)


class PublicReleaseAsset(Framework.TestCase):
    def setUp(self):
        self.authMode = "none"
        super().setUp()
        public_repo = self.g.get_repo("stellarcarbon/sorocarbon")
        self.release = public_repo.get_release(223668595)
        self.asset = self.release.assets[0]

    def testAttributes(self):
        self.assertEqual(self.release.id, 223668595)
        self.assertEqual(self.asset.id, 261582854)
        self.assertEqual(
            self.asset.url,
            "https://api.github.com/repos/stellarcarbon/sorocarbon/releases/assets/261582854",
        )
        self.assertEqual(self.asset.name, "sink-carbon_v0.3.0.wasm")
        self.assertEqual(self.asset.label, "")
        self.assertEqual(self.asset.content_type, "application/wasm")
        self.assertEqual(self.asset.state, "uploaded")
        self.assertEqual(self.asset.size, 4945)
        self.assertEqual(self.asset.digest, "sha256:2e4ca075988921c6c8bf01cc9e7bd4530830898f47346f1945cad930b7dfdf4d")
        self.assertEqual(
            self.asset.created_at,
            datetime(2025, 6, 6, 14, 43, 29, tzinfo=timezone.utc),
        )
        self.assertEqual(
            self.asset.updated_at,
            datetime(2025, 6, 6, 14, 43, 29, tzinfo=timezone.utc),
        )
        self.assertEqual(
            self.asset.browser_download_url,
            "https://github.com/stellarcarbon/sorocarbon/releases/download/v0.3.0__src_sink_carbon__sink-carbon_cli22.0.1/sink-carbon_v0.3.0.wasm",
        )
        self.assertEqual(self.asset.uploader.login, "github-actions[bot]")
        self.assertEqual(
            repr(self.asset),
            'GitReleaseAsset(url="https://api.github.com/repos/stellarcarbon/sorocarbon/releases/assets/261582854")',
        )

    def testDownload(self):
        with tempfile.TemporaryDirectory() as path:
            tmpf_path = path + "/asset"
            self.asset.download_asset(tmpf_path)
            with open(tmpf_path, "rb") as tmpf:
                tmpf.seek(0)
                hash = hashlib.sha256(tmpf.read())
                size = tmpf.tell()

        assert self.asset.digest
        digest = self.asset.digest.split(":")[1]
        self.assertEqual(hash.hexdigest(), digest)
        self.assertEqual(size, self.asset.size)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2024 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import github.GithubObject

from . import Framework


class ExposeAllAttributes(Framework.TestCase):
    def testAllClasses(self):
        authenticatedUser = self.g.get_user()
        namedUser = self.g.get_user("nvie")
        repository = authenticatedUser.get_repo("PyGithub")
        organization = self.g.get_organization("BeaverSoftware")
        plan = authenticatedUser.plan
        branch = repository.get_branch("master")
        commit = repository.get_commit("1292bf0e22c796e91cc3d6e24b544aece8c21f2a")
        commitStats = commit.stats
        commitStatus = commit.get_statuses()[0]
        milestone = repository.get_milestone(17)
        gist = self.g.get_gist("149016")
        gistComment = gist.get_comment(4565)
        gistFile = gist.files[".gitignore"]
        gistHistoryState = gist.history[0]
        gitCommit = repository.get_git_commit("be37b8a7f3a68631c32672dcd84d9eba27438ee6")
        gitAuthor = gitCommit.author
        gitTree = repository.get_git_tree("6f7c2d8c66d78863f7b91792deaead619799a1ce")
        gitTreeElement = gitTree.tree[0]
        gitBlob = repository.get_git_blob("681fb61f1761743a02f5c790f1c762cbfe8cfad1")
        gitRef = repository.get_git_ref("tags/v1.17.0")
        gitObject = gitRef.object
        issue = repository.get_issue(188)
        issueComment = issue.get_comment(22686536)
        issueEvent = issue.get_events()[0]
        issuePullRequest = issue.pull_request
        gitignoreTemplate = self.g.get_gitignore_template("Python")
        team = organization.get_team(141487)
        label = repository.get_label("Bug")
        pullRequest = repository.get_pull(31)
        pullRequestComment = pullRequest.get_review_comment(1580134)
        pullRequestPart = pullRequest.base
        file = pullRequest.get_files()[0]
        commitComment = repository.get_comment(3630301)
        rateLimit = self.g.get_rate_limit()
        hook = repository.get_hooks()[0]
        hookResponse = hook.last_response
        hookDescription = self.g.get_hooks()[0]
        comparison = repository.compare("master", "develop")
        contentFile = repository.get_contents("README.rst")
        permissions = repository.permissions
        event = repository.get_events()[0]
        notification = authenticatedUser.get_notification("8406712")
        notificationSubject = notification.subject

        missingAttributes = self.gatherMissingAttributes(
            [
                authenticatedUser,
                # authorization,  # Security issue if put as-is in ReplayData
                # authorizationApplication,  # Security issue if put as-is in ReplayData
                branch,
                commit,
                commitComment,
                commitStats,
                commitStatus,
                comparison,
                contentFile,
                # download,  # Deprecated: https://github.com/blog/1302-goodbye-uploads
                event,
                file,
                gist,
                gistComment,
                gistFile,
                gistHistoryState,
                gitAuthor,
                gitBlob,
                gitCommit,
                gitignoreTemplate,
                gitObject,
                gitRef,
                # gitTag,
                gitTree,
                gitTreeElement,
                hook,
                hookDescription,
                hookResponse,
                issue,
                issueComment,
                issueEvent,
                issuePullRequest,
                label,
                milestone,
                namedUser,
                notification,
                notificationSubject,
                organization,
                permissions,
                plan,
                pullRequest,
                pullRequestComment,
                # pullRequestMergeStatus,  # Only obtained when merging a pull request through the API
                pullRequestPart,
                rateLimit,
                repository,
                # repositoryKey,  # Security issue if put as-is in ReplayData
                # tag,
                team,
                # userKey,  # Security issue if put as-is in ReplayData
            ]
        )

        for className, attributesMissingInClass in sorted(missingAttributes.items()):
            for attrName, value in sorted(attributesMissingInClass.items()):
                print(className, attrName, "->", repr(value))

        self.assertEqual(sum(len(attrs) for attrs in missingAttributes.values()), 0)

    def findMissingAttributes(self, obj):
        if isinstance(obj, github.GithubObject.CompletableGithubObject):
            obj.update()
        className = obj.__class__.__name__
        missingAttributes = {}
        for attribute in obj.raw_data:
            if attribute != "_links":
                if not hasattr(obj, attribute):
                    missingAttributes[attribute] = obj.raw_data[attribute]
        return (className, missingAttributes)

    def gatherMissingAttributes(self, objs):
        allMissingAttributes = dict()

        for obj in objs:
            className, attributesMissingInClass = self.findMissingAttributes(obj)
            if len(attributesMissingInClass) > 0:
                if className not in allMissingAttributes:
                    allMissingAttributes[className] = dict()
                allMissingAttributes[className].update(attributesMissingInClass)

        return allMissingAttributes



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Jannis Gebauer <ja.geb@me.com>                                #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from datetime import datetime, timezone

from . import Framework


class GistComment(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.comment = self.g.get_gist("2729810").get_comment(323629)

    def testAttributes(self):
        self.assertIsNone(self.comment.author_association)
        self.assertEqual(self.comment.body, "Comment created by PyGithub")
        self.assertEqual(
            self.comment.created_at,
            datetime(2012, 5, 19, 7, 7, 57, tzinfo=timezone.utc),
        )
        self.assertEqual(self.comment.id, 323629)
        self.assertIsNone(self.comment.node_id)
        self.assertEqual(
            self.comment.updated_at,
            datetime(2012, 5, 19, 7, 7, 57, tzinfo=timezone.utc),
        )
        self.assertEqual(self.comment.url, "https://api.github.com/gists/2729810/comments/323629")
        self.assertEqual(self.comment.user.login, "jacquev6")
        self.assertEqual(
            repr(self.comment),
            'GistComment(user=NamedUser(login="jacquev6"), id=323629)',
        )

    def testLazyAttributes(self):
        comment = self.g.withLazy(True).get_gist("lazy").get_comment(42)
        self.assertEqual(str(comment), "GistComment(user=None, id=42)")
        self.assertEqual(comment.id, 42)
        self.assertEqual(comment.url, "/gists/lazy/comments/42")

    def testEdit(self):
        self.comment.edit("Comment edited by PyGithub")
        self.assertEqual(self.comment.body, "Comment edited by PyGithub")
        self.assertEqual(
            self.comment.updated_at,
            datetime(2012, 5, 19, 7, 12, 32, tzinfo=timezone.utc),
        )

    def testDelete(self):
        self.comment.delete()



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from . import Framework


class Issue33(Framework.TestCase):  # https://github.com/jacquev6/PyGithub/issues/33
    def setUp(self):
        super().setUp()
        self.repo = self.g.get_user("openframeworks").get_repo("openFrameworks")

    def testOpenIssues(self):
        # reduce the size of the record data file by executing the following:
        # cat -n tests/ReplayData/Issue33.testOpenIssues.txt | while read -r lineno line
        # do
        #   if [[ $(( lineno % 11 )) -eq 10 ]]
        #   then
        #     jq -c "[.[] | { id: .id }]" <<< "$line"
        #   else
        #     echo "$line"
        #   fi
        # done > tests/ReplayData/Issue33.testOpenIssues.txt.new
        # mv tests/ReplayData/Issue33.testOpenIssues.txt.new tests/ReplayData/Issue33.testOpenIssues.txt
        self.assertEqual(len(list(self.repo.get_issues())), 338)

    def testClosedIssues(self):
        # reduce the size of the record data file by executing the following:
        # cat -n tests/ReplayData/Issue33.testClosedIssues.txt | while read -r lineno line
        # do
        #   if [[ $(( lineno % 11 )) -eq 10 ]]
        #   then
        #     jq -c "[.[] | { id: .id }]" <<< "$line"
        #   else
        #     echo "$line"
        #   fi
        # done > tests/ReplayData/Issue33.testClosedIssues.txt.new
        # mv tests/ReplayData/Issue33.testClosedIssues.txt.new tests/ReplayData/Issue33.testClosedIssues.txt
        self.assertEqual(len(list(self.repo.get_issues(state="closed"))), 950)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2021 Amador Pahim <apahim@redhat.com>                              #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import github

from . import Framework

REPO_NAME = "PyGithub/PyGithub"


class PoolSize(Framework.TestCase):
    def setUp(self):
        self.setPoolSize(20)
        super().setUp()

    def testReturnsRepoAfterSettingPoolSize(self):
        repository = self.g.get_repo(REPO_NAME)
        self.assertIsInstance(repository, github.Repository.Repository)
        self.assertEqual(repository.full_name, REPO_NAME)

    def testReturnsRepoAfterSettingPoolSizeHttp(self):
        g = github.Github(
            auth=self.oauth_token,
            base_url="http://my.enterprise.com",
            pool_size=20,
        )
        repository = g.get_repo(REPO_NAME)
        self.assertIsInstance(repository, github.Repository.Repository)
        self.assertEqual(repository.full_name, REPO_NAME)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2024 Pasha Fateev <pasha@autokitteh.com>                           #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from datetime import datetime, timezone

from . import Framework


class Copilot(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.org_name = "BeaverSoftware"
        self.copilot = self.g.get_organization(self.org_name).get_copilot()

    def testAttributes(self):
        self.assertEqual(self.copilot.org_name, "BeaverSoftware")
        self.assertEqual(repr(self.copilot), 'Copilot(org_name="BeaverSoftware")')

        seats = list(self.copilot.get_seats())
        self.assertEqual(len(seats), 1)
        seat = seats[0]
        self.assertEqual(seat.created_at, datetime(2010, 7, 9, 6, 10, 6, tzinfo=timezone.utc))
        self.assertEqual(seat.updated_at, datetime(2012, 5, 26, 11, 25, 48, tzinfo=timezone.utc))
        self.assertEqual(seat.pending_cancellation_date, None)
        self.assertEqual(seat.last_activity_at, datetime(2012, 5, 26, 14, 59, 39, tzinfo=timezone.utc))
        self.assertEqual(seat.last_activity_editor, "vscode/1.0.0")
        self.assertEqual(seat.organization.login, self.org_name)
        self.assertEqual(seat.plan_type, "business")
        self.assertEqual(seat.assignee.login, "pashafateev")
        self.assertEqual(repr(seat), 'CopilotSeat(assignee=NamedUser(login="pashafateev"))')

    def testGetSeats(self):
        seats = self.copilot.get_seats()
        self.assertListKeyEqual(seats, lambda s: s.assignee.login, ["pashafateev"])

    def testAddSeats(self):
        seats_created = self.copilot.add_seats(["pashafateev"])
        self.assertEqual(seats_created, 1)

    def testRemoveSeats(self):
        seats_cancelled = self.copilot.remove_seats(["pashafateev"])
        self.assertEqual(seats_cancelled, 1)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Jannis Gebauer <ja.geb@me.com>                                #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2017 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Danilo Martins <mawkee@gmail.com>                             #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2024 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2024 Kian-Meng Ang <kianmeng.ang@gmail.com>                        #
# Copyright 2024 iarspider <iarspider@gmail.com>                               #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from . import Framework


class Commit(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.commit = self.g.get_user().get_repo("PyGithub").get_commit("1292bf0e22c796e91cc3d6e24b544aece8c21f2a")
        self.commit.author.login  # to force lazy completion

    def testAttributes(self):
        self.assertEqual(self.commit.author.login, "jacquev6")
        self.assertEqual(
            self.commit.comments_url,
            "https://api.github.com/repos/jacquev6/PyGithub/commits/1292bf0e22c796e91cc3d6e24b544aece8c21f2a/comments",
        )
        self.assertEqual(
            self.commit.commit.url,
            "https://api.github.com/repos/jacquev6/PyGithub/git/commits/1292bf0e22c796e91cc3d6e24b544aece8c21f2a",
        )
        self.assertEqual(self.commit.committer.login, "jacquev6")
        self.assertEqual(len(list(self.commit.files)), 1)
        self.assertEqual(self.commit.files.totalCount, 1)
        self.assertEqual(self.commit.files[0].additions, 0)
        self.assertEqual(
            self.commit.files[0].blob_url,
            "https://github.com/jacquev6/PyGithub/blob/1292bf0e22c796e91cc3d6e24b544aece8c21f2a/github%2FGithubObjects%2FGitAuthor.py",
        )
        self.assertEqual(self.commit.files[0].changes, 20)
        self.assertEqual(self.commit.files[0].deletions, 20)
        self.assertEqual(self.commit.files[0].filename, "github/GithubObjects/GitAuthor.py")
        self.assertTrue(isinstance(self.commit.files[0].patch, str))
        self.assertEqual(
            self.commit.files[0].raw_url,
            "https://github.com/jacquev6/PyGithub/raw/1292bf0e22c796e91cc3d6e24b544aece8c21f2a/github%2FGithubObjects%2FGitAuthor.py",
        )
        self.assertEqual(self.commit.files[0].sha, "ca6a3c616fc1367b6d01d04a7cf6ee27cf216f26")
        self.assertEqual(self.commit.files[0].status, "modified")
        self.assertEqual(
            self.commit.html_url, "https://github.com/jacquev6/PyGithub/commit/1292bf0e22c796e91cc3d6e24b544aece8c21f2a"
        )
        self.assertEqual(
            self.commit.node_id, "MDY6Q29tbWl0NDQ2MzY1NzM1OjEyOTJiZjBlMjJjNzk2ZTkxY2MzZDZlMjRiNTQ0YWVjZThjMjFmMmE="
        )
        self.assertEqual(len(self.commit.parents), 1)
        self.assertEqual(self.commit.parents[0].sha, "b46ed0dfde5ad02d3b91eb54a41c5ed960710eae")
        self.assertIsNone(self.commit.repository)
        self.assertEqual(self.commit.sha, "1292bf0e22c796e91cc3d6e24b544aece8c21f2a")
        self.assertEqual(self.commit.stats.deletions, 20)
        self.assertEqual(self.commit.stats.additions, 0)
        self.assertEqual(self.commit.stats.total, 20)
        self.assertIsNone(self.commit.text_matches)
        self.assertEqual(
            self.commit.url,
            "https://api.github.com/repos/jacquev6/PyGithub/commits/1292bf0e22c796e91cc3d6e24b544aece8c21f2a",
        )
        self.assertEqual(self.commit.commit.tree.sha, "4c6bd50994f0f9823f898b1c6c964ad7d4fa11ab")
        self.assertEqual(repr(self.commit), 'Commit(sha="1292bf0e22c796e91cc3d6e24b544aece8c21f2a")')

    def testLazyAttributes(self):
        commit = self.g.withLazy(True).get_repo("lazy/repo").get_commit("sha")
        self.assertEqual(str(commit), 'Commit(sha="sha")')
        self.assertEqual(commit._identity, "sha")
        self.assertEqual(commit.sha, "sha")
        self.assertEqual(commit.url, "/repos/lazy/repo/commits/sha")

    def testGetBranchesWhereHead(self):
        repo = self.g.get_repo("PyGithub/PyGithub")
        commit = repo.get_commit("0791cc7b1a706ab5d7c607ddff35de4d486ba3e9")
        self.assertListKeyEqual(
            commit.get_branches_where_head(),
            lambda b: b.name,
            ["release-v2-0"],
        )

    def testGetComments(self):
        self.assertListKeyEqual(
            self.commit.get_comments(),
            lambda c: c.id,
            [1347033, 1347083, 1347397, 1349654],
        )

    def testCreateComment(self):
        comment = self.commit.create_comment("Comment created by PyGithub")
        self.assertEqual(comment.id, 1361949)
        self.assertEqual(comment.line, None)
        self.assertEqual(comment.path, None)
        self.assertEqual(comment.position, None)

    def testCreateCommentOnFileLine(self):
        comment = self.commit.create_comment(
            "Comment created by PyGithub",
            path="codegen/templates/GithubObject.MethodBody.UseResult.py",
            line=26,
        )
        self.assertEqual(comment.id, 1362000)
        self.assertEqual(comment.line, 26)
        self.assertEqual(comment.path, "codegen/templates/GithubObject.MethodBody.UseResult.py")
        self.assertEqual(comment.position, None)

    def testCreateCommentOnFilePosition(self):
        comment = self.commit.create_comment(
            "Comment also created by PyGithub",
            path="codegen/templates/GithubObject.MethodBody.UseResult.py",
            position=3,
        )
        self.assertEqual(comment.id, 1362001)
        self.assertEqual(comment.line, None)
        self.assertEqual(comment.path, "codegen/templates/GithubObject.MethodBody.UseResult.py")
        self.assertEqual(comment.position, 3)

    def testCreateStatusWithoutOptionalParameters(self):
        status = self.commit.create_status("pending")
        self.assertEqual(status.id, 277031)
        self.assertEqual(status.state, "pending")
        self.assertEqual(status.target_url, None)
        self.assertEqual(status.description, None)

    def testCreateStatusWithAllParameters(self):
        status = self.commit.create_status(
            "success",
            "https://github.com/jacquev6/PyGithub/issues/67",
            "Status successfully created by PyGithub",
        )
        self.assertEqual(status.id, 277040)
        self.assertEqual(status.state, "success")
        self.assertEqual(status.target_url, "https://github.com/jacquev6/PyGithub/issues/67")
        self.assertEqual(status.description, "Status successfully created by PyGithub")

    def testGetPulls(self):
        commit = self.g.get_user().get_repo("PyGithub").get_commit("e44d11d565c022496544dd6ed1f19a8d718c2b0c")
        self.assertListKeyEqual(commit.get_pulls(), lambda c: c.number, [1431])



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 Shinichi TAMURA <shnch.tmr@gmail.com>                         #
# Copyright 2018 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2024 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2024 Min RK <benjaminrk@gmail.com>                                 #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from datetime import datetime, timezone

from urllib3.exceptions import InsecureRequestWarning

import github
from github import Consts
from github.Auth import AppAuth, AppInstallationAuth

from . import Framework, GithubIntegration


class Installation(Framework.BasicTestCase):
    def setUp(self):
        super().setUp()
        app_id = 36541767
        private_key = GithubIntegration.PRIVATE_KEY
        self.auth = AppAuth(app_id, private_key)
        self.integration = github.GithubIntegration(auth=self.auth)
        self.installations = list(self.integration.get_installations())
        self.installation = self.installations[0]

    def testAttributes(self):
        self.assertEqual(
            self.installation.access_tokens_url, "https://api.github.com/app/installations/36541767/access_tokens"
        )
        self.assertEqual(self.installation.account.login, "EnricoMi")
        self.assertEqual(self.installation.app_id, 319953)
        self.assertEqual(self.installation.app_slug, "publish-test-results")
        self.assertIsNone(self.installation.client_id)
        self.assertIsNone(self.installation.contact_email)
        self.assertEqual(self.installation.created_at, datetime(2023, 4, 17, 16, 18, 5, tzinfo=timezone.utc))
        self.assertEqual(self.installation.events, [])
        self.assertEqual(self.installation.has_multiple_single_files, False)
        self.assertEqual(self.installation.html_url, "https://github.com/settings/installations/36541767")
        self.assertEqual(self.installation.id, 36541767)
        self.assertEqual(
            self.installation.permissions,
            {"checks": "write", "issues": "read", "contents": "read", "metadata": "read", "pull_requests": "write"},
        )
        self.assertEqual(self.installation.repositories_url, "https://api.github.com/installation/repositories")
        self.assertEqual(self.installation.repository_selection, "selected")
        self.assertIsNone(self.installation.single_file_name)
        self.assertEqual(self.installation.single_file_paths, [])
        self.assertIsNone(self.installation.suspended_at)
        self.assertIsNone(self.installation.suspended_by)
        self.assertEqual(self.installation.target_id, 44700269)
        self.assertEqual(self.installation.target_type, "User")
        self.assertEqual(self.installation.updated_at, datetime(2023, 6, 8, 7, 38, 12, tzinfo=timezone.utc))

    def testGetRepos(self):
        repos = list(self.installation.get_repos())
        self.assertEqual(len(repos), 2)
        self.assertListEqual([repo.full_name for repo in repos], ["EnricoMi/sandbox", "EnricoMi/python"])

    def testGetGithubForInstallation(self):
        # with verify=False, urllib3.connectionpool rightly may issue an InsecureRequestWarning
        # we ignore InsecureRequestWarning from urllib3.connectionpool
        with self.ignoreWarning(category=InsecureRequestWarning, module="urllib3.connectionpool"):
            kwargs = dict(
                auth=AppAuth(319953, GithubIntegration.PRIVATE_KEY),
                # http protocol used to deviate from default base url, recording data might require https
                base_url="http://api.github.com",
                timeout=Consts.DEFAULT_TIMEOUT + 10,
                user_agent="PyGithub/Python-Test",
                per_page=Consts.DEFAULT_PER_PAGE + 10,
                verify=False,
                retry=3,
                pool_size=10,
                seconds_between_requests=100,
                seconds_between_writes=1000,
                # v3: this should not be the default value, so if this has been changed in v3,
                # change it here is well
                lazy=True,
            )

            # assert kwargs consists of ALL requester constructor arguments
            self.assertEqual(kwargs.keys(), github.Requester.Requester.__init__.__annotations__.keys())

            integration = github.GithubIntegration(**kwargs)
            installations = list(integration.get_installations())
            installation = installations[0]

            g = installation.get_github_for_installation()

            self.assertIsInstance(g._Github__requester.auth, AppInstallationAuth)

            actual = g._Github__requester.kwargs
            kwargs.update(auth=str(AppInstallationAuth))
            actual.update(auth=str(type(actual["auth"])))
            self.assertDictEqual(kwargs, actual)

            repo = g.get_repo("PyGithub/PyGithub")
            self.assertEqual(repo.id, 3544490)
            self.assertEqual(repo.full_name, "PyGithub/PyGithub")

    def testRequester(self):
        self.assertEqual(len(self.installations), 1)
        installation = self.installations[0]
        assert installation.requester is installation._requester



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2018 Shubham Singh <41840111+singh811@users.noreply.github.com>    #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 TechnicalPirate <35609336+TechnicalPirate@users.noreply.github.com>#
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Christoph Reiter <reiter.christoph@gmail.com>                 #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

from __future__ import annotations

from datetime import datetime, timedelta, timezone

import github

from . import Framework


class Migration(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.user = self.g.get_user()
        self.migration = self.user.get_migrations()[0]

    def testAttributes(self):
        self.assertIsNone(self.migration.archive_url)
        self.assertEqual(
            self.migration.created_at, datetime(2018, 9, 14, 1, 35, 35, tzinfo=timezone(timedelta(seconds=19800)))
        )
        self.assertEqual(self.migration.exclude, ["repositories"])
        self.assertEqual(self.migration.exclude_attachments, False)
        self.assertEqual(self.migration.exclude_git_data, False)
        self.assertEqual(self.migration.exclude_metadata, False)
        self.assertEqual(self.migration.exclude_owner_projects, False)
        self.assertEqual(self.migration.exclude_releases, False)
        self.assertEqual(self.migration.guid, "608bceae-b790-11e8-8b43-4e3cb0dd56cc")
        self.assertEqual(self.migration.id, 25320)
        self.assertEqual(self.migration.lock_repositories, False)
        self.assertEqual(self.migration.node_id, "MDk6TWlncmF0aW9uMjUzMjA=")
        self.assertEqual(self.migration.org_metadata_only, False)
        self.assertEqual(self.migration.owner.login, "singh811")
        self.assertEqual(self.migration.guid, "608bceae-b790-11e8-8b43-4e3cb0dd56cc")
        self.assertEqual(self.migration.repositories[0].full_name, "singh811/sample-repo")
        self.assertEqual(self.migration.state, "exported")
        self.assertEqual(self.migration.lock_repositories, False)
        self.assertEqual(self.migration.exclude_attachments, False)
        self.assertEqual(len(self.migration.repositories), 1)
        self.assertEqual(self.migration.repositories[0].name, "sample-repo")
        self.assertEqual(
            self.migration.updated_at, datetime(2018, 9, 14, 1, 35, 46, tzinfo=timezone(timedelta(seconds=19800)))
        )
        self.assertEqual(self.migration.url, "https://api.github.com/user/migrations/25320")
        self.assertEqual(
            self.migration.created_at,
            datetime(2018, 9, 14, 1, 35, 35, tzinfo=timezone(timedelta(seconds=19800))),
        )
        self.assertEqual(
            self.migration.updated_at,
            datetime(2018, 9, 14, 1, 35, 46, tzinfo=timezone(timedelta(seconds=19800))),
        )
        self.assertEqual(
            repr(self.migration),
            'Migration(url="https://api.github.com/user/migrations/25320", state="exported")',
        )

    def testGetArchiveUrlWhenNotExported(self):
        self.assertRaises(github.UnknownObjectException, lambda: self.migration.get_archive_url())

    def testGetStatus(self):
        self.assertEqual(self.migration.get_status(), "exported")

    def testGetArchiveUrlWhenExported(self):
        self.assertEqual(
            self.migration.get_archive_url(),
            "https://github-cloud.s3.amazonaws.com/migration/25320/24575?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=AKIAISTNZFOVBIJMK3TQ%2F20180913%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Date=20180913T201100Z&X-Amz-Expires=300&X-Amz-Signature=a0aeb638facd0c78c1ed3ca86022eddbee91e5fe1bb48ee830f54b8b7b305026&X-Amz-SignedHeaders=host&actor_id=41840111&response-content-disposition=filename%3D608bceae-b790-11e8-8b43-4e3cb0dd56cc.tar.gz&response-content-type=application%2Fx-gzip",
        )

    def testDelete(self):
        self.assertEqual(self.migration.delete(), None)

    def testGetArchiveUrlWhenDeleted(self):
        self.assertRaises(github.UnknownObjectException, lambda: self.migration.get_archive_url())

    def testUnlockRepo(self):
        self.assertEqual(self.migration.unlock_repo("sample-repo"), None)



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2019 Adam Baratz <adam.baratz@gmail.com>                           #
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################


from __future__ import annotations

from datetime import datetime, timezone
from operator import attrgetter

from . import Framework


class Topic(Framework.TestCase):
    def setUp(self):
        super().setUp()
        self.topics = list(self.g.search_topics("python"))

    def testAttributes(self):
        topic = self.topics[0]
        self.assertIsNone(topic.aliases)
        self.assertIsNone(topic.logo_url)

        self.assertEqual(topic.name, "python")
        self.assertEqual(topic.display_name, "Python")
        self.assertIsNone(topic.related)
        self.assertIsNone(topic.repository_count)
        self.assertEqual(
            topic.short_description,
            "Python is a dynamically typed programming language.",
        )
        self.assertEqual(
            topic.description,
            "Python is a dynamically-typed garbage-collected programming language "
            "developed by Guido van Rossum in the late 80s to replace ABC. Much like the "
            "programming language Ruby, Python was designed to be easily read by "
            "programmers. Because of its large following and many libraries, Python can "
            "be implemented and used to do anything from webpages to scientific research.",
        )
        self.assertEqual(topic.created_by, "Guido van Rossum")
        self.assertEqual(topic.released, "February 20, 1991")
        self.assertEqual(topic.created_at, datetime(2016, 12, 7, 0, 7, 2, tzinfo=timezone.utc))
        self.assertIsNone(topic.text_matches)
        self.assertEqual(topic.updated_at, datetime(2025, 1, 7, 9, 22, 52, tzinfo=timezone.utc))
        self.assertEqual(topic.featured, True)
        self.assertEqual(topic.curated, True)
        self.assertEqual(topic.score, 1.0)

        self.assertEqual(topic.__repr__(), 'Topic(name="python")')

    def testNamesFromSearchResults(self):
        expected_names = [
            "python",
            "django",
            "hacktoberfest",
            "flask",
            "python-script",
            "opencv-python",
            "numpy",
            "backend",
            "fastapi",
            "ruby",
            "python-library",
            "keras",
            "selenium-python",
            "tkinter-python",
            "scikit-learn",
            "pandas-python",
            "streamlit",
            "leetcode-python",
            "python36",
            "machinelearning-python",
            "nltk-python",
            "data-analysis-python",
            "micropython",
            "python37",
            "langchain-python",
            "learning-python",
            "oops-in-python",
            "python-telegram-bot",
            "python-package",
            "python-project",
        ]
        self.assertListKeyEqual(self.topics, attrgetter("name"), expected_names)



#!/usr/bin/env python
############################ Copyrights and license ############################
#                                                                              #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
# Copyright 2019 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2019 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2020 Steve Kowalik <steven@wedontsleep.org>                        #
# Copyright 2020 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2023 Enrico Minack <github@enrico.minack.dev>                      #
# Copyright 2023 Jirka Borovec <6035284+Borda@users.noreply.github.com>        #
# Copyright 2023 Jonathan Leitschuh <jonathan.leitschuh@gmail.com>             #
# Copyright 2024 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import os
import subprocess

eightySharps = "#" * 80


def generateLicenseSection(filename):
    yield "############################ Copyrights and license ############################"
    yield "#                                                                              #"
    for year, name in sorted(listContributors(filename)):
        line = "# Copyright " + year + " " + name
        line += (79 - len(line)) * " " + "#"
        yield line
    yield "#                                                                              #"
    yield "# This file is part of PyGithub.                                               #"
    yield "# http://pygithub.readthedocs.io/                                              #"
    yield "#                                                                              #"
    yield "# PyGithub is free software: you can redistribute it and/or modify it under    #"
    yield "# the terms of the GNU Lesser General Public License as published by the Free  #"
    yield "# Software Foundation, either version 3 of the License, or (at your option)    #"
    yield "# any later version.                                                           #"
    yield "#                                                                              #"
    yield "# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #"
    yield "# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #"
    yield "# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #"
    yield "# details.                                                                     #"
    yield "#                                                                              #"
    yield "# You should have received a copy of the GNU Lesser General Public License     #"
    yield "# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #"
    yield "#                                                                              #"
    yield "################################################################################"


def listContributors(filename):
    contributors = set()
    result = subprocess.check_output(
        ["git", "log", "--follow", "--format=format:%ad %an <%ae>", "--date=short", "--", filename],
        text=True,
    )
    for line in result.split("\n"):
        year = line[0:4]
        name = line[11:]
        contributors.add((year, name))
    return contributors


def extractBodyLines(lines):
    bodyLines = []

    seenEndOfHeader = False

    for line in lines:
        if len(line) > 0 and line[0] != "#":
            seenEndOfHeader = True
        if seenEndOfHeader:
            bodyLines.append(line)
        # else:
        #     print "HEAD:", line
        if line == eightySharps:
            seenEndOfHeader = True

    # print "BODY:", "\nBODY: ".join(bodyLines)

    return bodyLines


class PythonHeader:
    def fix(self, filename, lines):
        isExecutable = len(lines) > 0 and lines[0].startswith("#!")
        newLines = []

        if isExecutable:
            newLines.append("#!/usr/bin/env python")

        for line in generateLicenseSection(filename):
            newLines.append(line)

        bodyLines = extractBodyLines(lines)

        if len(bodyLines) > 0 and bodyLines[0] != "":
            newLines.append("")
            if "import " not in bodyLines[0] and bodyLines[0] != '"""' and not bodyLines[0].startswith("##########"):
                newLines.append("")
        newLines += bodyLines

        return newLines


class StandardHeader:
    def fix(self, filename, lines):
        newLines = []

        for line in generateLicenseSection(filename):
            newLines.append(line)

        bodyLines = extractBodyLines(lines)

        if len(bodyLines) > 0 and bodyLines[0] != "":
            newLines.append("")
        newLines += bodyLines

        return newLines


def findHeadersAndFiles():
    for root, dirs, files in os.walk(".", topdown=True):
        for dir in list(dirs):
            if dir.startswith("."):
                dirs.remove(dir)
        for excluded in ["developer.github.com", "build", "venv", "PyGithub.egg-info", "requirements", "pre-commit"]:
            if excluded in dirs:
                dirs.remove(excluded)

        for filename in files:
            fullname = os.path.join(root, filename)
            if filename == "GithubCredentials.py":
                pass
            elif filename.endswith(".py"):
                yield (PythonHeader(), fullname)
            elif filename in ["COPYING", "COPYING.LESSER", "MAINTAINERS"]:
                pass
            elif filename.endswith(".rst") or filename.endswith(".md"):
                pass
            elif filename == ".gitignore":
                yield (StandardHeader(), fullname)
            elif "ReplayData" in fullname:
                pass
            elif fullname.endswith(".pyi"):
                pass
            elif fullname.endswith(".pyc"):
                pass
            else:
                print(f"Don't know what to do with {filename} in {root}")


def main():
    for header, filename in findHeadersAndFiles():
        print("Analyzing", filename)
        with open(filename, encoding="utf-8") as f:
            lines = list(line.rstrip() for line in f)
        newLines = header.fix(filename, lines)
        if newLines != lines:
            print(" => actually modifying", filename)
            with open(filename, "w", encoding="utf-8") as f:
                for line in newLines:
                    f.write(line + "\n")


if __name__ == "__main__":
    main()



############################ Copyrights and license ############################
#                                                                              #
# Copyright 2025 Enrico Minack <github@enrico.minack.dev>                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import argparse
import difflib
import sys
from typing import Union

import libcst as cst
from libcst import SimpleWhitespace


class SingleLineStatementTransformer(cst.CSTTransformer):
    def __init__(self, function: str):
        super().__init__()
        self.function = function
        self.in_function = False

    def visit_FunctionDef(self, node: cst.FunctionDef):
        if node.name.value == self.function:
            self.in_function = True

    def leave_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef):
        self.in_function = False
        return updated_node

    def leave_Call(self, original_node: cst.Call, updated_node: cst.Call) -> cst.BaseExpression:
        if self.in_function:
            return updated_node.with_changes(
                whitespace_after_func=SimpleWhitespace(""), whitespace_before_args=SimpleWhitespace("")
            )
        return updated_node

    def leave_Arg(
        self, original_node: cst.Arg, updated_node: cst.Arg
    ) -> Union[cst.Arg, cst.FlattenSentinel[cst.Arg], cst.RemovalSentinel]:
        if self.in_function:
            return updated_node.with_changes(
                whitespace_after_star=SimpleWhitespace(""), whitespace_after_arg=SimpleWhitespace("")
            )
        return updated_node

    def leave_LeftParen(self, original_node: cst.LeftParen, updated_node: cst.LeftParen) -> cst.LeftParen:
        if self.in_function:
            return updated_node.with_changes(whitespace_after=SimpleWhitespace(""))
        return updated_node

    def leave_RightParen(self, original_node: cst.RightParen, updated_node: cst.RightParen) -> cst.RightParen:
        if self.in_function:
            return updated_node.with_changes(whitespace_before=SimpleWhitespace(""))
        return updated_node

    def leave_LeftCurlyBrace(
        self, original_node: cst.LeftCurlyBrace, updated_node: cst.LeftCurlyBrace
    ) -> cst.LeftCurlyBrace:
        if self.in_function:
            return updated_node.with_changes(whitespace_after=SimpleWhitespace(""))
        return updated_node

    def leave_RightCurlyBrace(
        self, original_node: cst.RightCurlyBrace, updated_node: cst.RightCurlyBrace
    ) -> cst.RightCurlyBrace:
        if self.in_function:
            return updated_node.with_changes(whitespace_before=SimpleWhitespace(""))
        return updated_node

    def leave_LeftSquareBracket(
        self, original_node: cst.LeftSquareBracket, updated_node: cst.LeftSquareBracket
    ) -> cst.LeftSquareBracket:
        if self.in_function:
            return updated_node.with_changes(whitespace_after=SimpleWhitespace(""))
        return updated_node

    def leave_RightSquareBracket(
        self, original_node: cst.RightSquareBracket, updated_node: cst.RightSquareBracket
    ) -> cst.RightSquareBracket:
        if self.in_function:
            return updated_node.with_changes(whitespace_before=SimpleWhitespace(""))
        return updated_node

    def leave_Comma(self, original_node: cst.Comma, updated_node: cst.Comma) -> Union[cst.Comma, cst.MaybeSentinel]:
        if self.in_function:
            return updated_node.with_changes(
                whitespace_before=SimpleWhitespace(""), whitespace_after=SimpleWhitespace(" ")
            )
        return updated_node


def main(filename: str, function: str, dry_run: bool) -> bool:
    with open(filename) as r:
        code = "".join(r.readlines())

    tree = cst.parse_module(code)
    transformer = SingleLineStatementTransformer(function)
    tree_updated = tree.visit(transformer)

    if dry_run:
        diff = "".join(difflib.unified_diff(code.splitlines(1), tree_updated.code.splitlines(1)))
        if diff:
            print(f"Diff of {filename}:")
            print(diff)
            print()
            return True
    else:
        if not tree_updated.deep_equals(tree):
            with open(filename, "w") as w:
                w.write(tree_updated.code)
            return True

    return False


def parse_args():
    args_parser = argparse.ArgumentParser(
        description="Removes newlines from all statements in a specific function of one test file"
    )
    args_parser.add_argument("filename", help="Path of test file")
    args_parser.add_argument("function", help="Name of the test function")
    args_parser.add_argument(
        "--dry-run", default=False, action="store_true", help="Show prospect changes and do not modify the file"
    )
    args_parser.add_argument(
        "--exit-code", default=False, action="store_true", help="Indicate changes via non-zero exit code"
    )

    if len(sys.argv) == 1:
        args_parser.print_help()
        sys.exit(1)
    return args_parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    changed = main(args.filename, args.function, args.dry_run)
    if args.exit_code and changed:
        sys.exit(1)
